# -*- coding: utf-8 -*-
"""
Tests for utilities related to python structures such as lists.
"""
import pytest

from nessai.utils.structures import replace_in_list


def test_replace_in_list():
    """
    Test if the list produced contains the correct entries in the correct
    locations
    """
    x = [1, 2, 3]
    replace_in_list(x, [1, 2], [5, 4])
    assert x == [5, 4, 3]


def test_replace_in_list_item():
    """
    Test if items are correctly converted to lists in replace_in_list function
    """
    x = [1, 2, 3]
    replace_in_list(x, 3, 4)
    assert x == [1, 2, 4]


def test_replace_in_list_single():
    """Test the function with a list of strings"""
    x = ["aa", "bb", "cc"]
    replace_in_list(x, ["aa", "bb"], ["dd", "ee"])
    assert x == ["dd", "ee", "cc"]


def test_replace_in_list_single_str():
    """Test the function with single strings"""
    x = ["aa", "bb", "cc"]
    replace_in_list(x, "aa", "dd")
    assert x == ["dd", "bb", "cc"]


def test_different_lengths():
    """
    Assert an error is raised if the targets and replacements are different
    lengths.
    """
    with pytest.raises(RuntimeError) as excinfo:
        replace_in_list([1, 2], [1, 2], 3)
    assert "Targets and replacements are different lengths!" in str(
        excinfo.value
    )


def test_missing_targets():
    """Assert an error is raised if a target is not in the target list."""
    with pytest.raises(ValueError) as excinfo:
        replace_in_list([1, 2], 4, 3)
    assert "Targets [4] not in list: [1, 2]" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test the Rescale class.
"""
import numpy as np
import pytest
from unittest.mock import create_autospec

from nessai.livepoint import numpy_array_to_live_points
from nessai.reparameterisations import Rescale
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture()
def reparam():
    return create_autospec(Rescale)


@pytest.mark.parametrize("scale", [2, 2.0, [1, 2], {"x": 1, "y": 2}])
def test_init(scale):
    """Test the init method with different input types"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    reparam = Rescale(
        parameters=parameters, scale=scale, prior_bounds=prior_bounds
    )

    assert not set(reparam.scale.keys()) - set(parameters)
    assert isinstance(reparam.scale["x"], float)


@pytest.mark.parametrize("n", [1, 2])
def test_reparameterise(reparam, n):
    """Test the reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    x = numpy_array_to_live_points(np.ones((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.zeros((n, 2)), reparam.prime_parameters
    )
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = Rescale.reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x, x_out)
    np.testing.assert_array_equal(log_j_out, -np.log(8 * np.ones(n)))
    assert (x_prime_out["x_prime"] == -0.5).all()
    assert (x_prime_out["y_prime"] == 0.25).all()


@pytest.mark.parametrize("scale", [1e60, 1e-60])
def test_reparameterise_overflow(reparam, scale):
    """Test the reparameterise method with very small and large scales.

    Checks precision to 14 decimal places.
    """
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.scale = {"x": scale}
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(
        scale * x_array[:, np.newaxis], reparam.parameters
    )
    x_prime = numpy_array_to_live_points(
        np.ones((x_array.size, 1)), reparam.prime_parameters
    )
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = Rescale.reparameterise(
        reparam, x, x_prime, log_j
    )

    np.testing.assert_array_almost_equal(
        x_array, x_prime_out["x_prime"], decimal=14
    )
    assert (log_j == -np.log(scale)).all()


@pytest.mark.parametrize("n", [1, 2])
def test_inverse_reparameterise(reparam, n):
    """Test the inverse reparameterise method"""
    reparam.parameters = ["x", "y"]
    reparam.prime_parameters = ["x_prime", "y_prime"]
    reparam.scale = {"x": -2.0, "y": 4.0}
    x = numpy_array_to_live_points(np.zeros((n, 2)), reparam.parameters)
    x_prime = numpy_array_to_live_points(
        np.ones((n, 2)), reparam.prime_parameters
    )
    x_prime["x_prime"] *= -1
    log_j = np.zeros(n)

    x_out, x_prime_out, log_j_out = Rescale.inverse_reparameterise(
        reparam, x, x_prime, log_j
    )

    assert_structured_arrays_equal(x_prime, x_prime_out)
    np.testing.assert_array_equal(log_j_out, np.log(8 * np.ones(n)))
    assert (x_out["x"] == 2.0).all()
    assert (x_out["y"] == 4.0).all()


@pytest.mark.parametrize("scale", [1e60, 1e-60])
def test_inverse_reparameterise_overflow(reparam, scale):
    """
    Test the inverse_reparameterise method with very small and large scales.
    """
    reparam.parameters = ["x"]
    reparam.prime_parameters = ["x_prime"]
    reparam.scale = {"x": scale}
    x_array = np.arange(100.0, dtype=float)
    x = numpy_array_to_live_points(
        np.ones((x_array.size, 1)), reparam.parameters
    )
    x_prime = numpy_array_to_live_points(
        x_array[:, np.newaxis], reparam.prime_parameters
    )
    log_j = np.zeros(x.size)

    x_out, x_prime_out, log_j_out = Rescale.inverse_reparameterise(
        reparam, x, x_prime, log_j
    )

    np.testing.assert_array_equal(x_array * scale, x_out["x"])
    assert (log_j == np.log(scale)).all()


def test_init_no_scale():
    """Make sure an error is raised if the scale is not given"""
    with pytest.raises(RuntimeError) as excinfo:
        Rescale(scale=None)
    assert "Must specify a scale!" in str(excinfo.value)


@pytest.mark.parametrize("scale", [[1], [1, 2, 3]])
def test_init_incorrect_scale_list(scale):
    """Make sure an error is raised if the scale is the incorrect length"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        Rescale(parameters=parameters, scale=scale, prior_bounds=prior_bounds)

    assert "different length" in str(excinfo.value)


@pytest.mark.parametrize("scale", [{"x": 1}, {"x": 1, "y": 1, "z": 1}])
def test_init_incorrect_scale_dict(scale):
    """Make sure an error is raised if the scale keys to not match the \
            parameters.
    """
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(RuntimeError) as excinfo:
        Rescale(parameters=parameters, scale=scale, prior_bounds=prior_bounds)

    assert "Mismatched parameters" in str(excinfo.value)


def test_init_incorrect_scale_type():
    """Make sure an error is raised if the scale is the incorrect type"""
    parameters = ["x", "y"]
    prior_bounds = {"x": [-1, 1], "y": [-1, 1]}

    with pytest.raises(TypeError) as excinfo:
        Rescale(parameters=parameters, scale="1", prior_bounds=prior_bounds)

    assert "Scale input must be" in str(excinfo.value)



import sys

from numpy.random import seed
import numpy as np
import pytest
from scipy.stats import norm
import time
import torch
import multiprocessing

from nessai.model import Model


seed(170817)
torch.manual_seed(170817)

_requires_dependency_cache = dict()


@pytest.fixture()
def model():
    class TestModel(Model):
        def __init__(self):
            self.bounds = {"x": [-5, 5], "y": [-5, 5]}
            self.names = ["x", "y"]

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x), dtype="float")
            for n in self.names:
                log_p -= self.bounds[n][1] - self.bounds[n][0]
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

    return TestModel()


@pytest.fixture()
def flow_config():
    d = dict(
        max_epochs=5,
        model_config=dict(
            n_blocks=2,
            n_neurons=2,
            n_layers=1,
            device_tag="cpu",
            kwargs=dict(batch_norm_between_layers=False),
        ),
    )
    return d


@pytest.fixture()
def wait():
    """Sleep for 0.01s for timing tests.

    Time resolution on Windows is less than Linux and macOS
    """

    def func(*args, **kwargs):
        time.sleep(0.01)

    return func


@pytest.fixture(params=["fork", "spawn"])
def mp_context(request):
    """Multiprocessing context to test"""
    if request.param == "spawn":
        pytest.skip(
            "nessai does not currently support multiprocessing with the "
            "'spawn' method."
        )
    if sys.platform == "win32":
        pytest.skip("Windows does not support 'fork' method")
    return multiprocessing.get_context(request.param)


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "cuda: mark test to indicate it requires CUDA"
    )
    config.addinivalue_line(
        "markers",
        "requires(package): mark test to only run if the package can be "
        "imported",
    )
    config.addinivalue_line(
        "markers",
        "skip_on_windows: mark test to indicated it should be skipped on "
        "Windows",
    )


def pytest_runtest_setup(item):
    for mark in item.iter_markers(name="cuda"):
        if not torch.cuda.is_available():
            pytest.skip("Test requires CUDA")
    for mark in item.iter_markers(name="requires"):
        name = mark.args[0]
        if name in _requires_dependency_cache:
            skip_it = _requires_dependency_cache[name]
        else:
            try:
                __import__(name)
                skip_it = False
            except ImportError:
                skip_it = True

            _requires_dependency_cache[name] = skip_it

        reason = "Missing dependency: {}".format(name)
        if skip_it:
            pytest.skip(reason)
    for mark in item.iter_markers(name="skip_on_windows"):
        if sys.platform == "win32":
            pytest.skip("Test does not run on Windows")



# -*- coding: utf-8 -*-
"""
Functions related to computing the posterior samples.
"""
import numpy as np

from .evidence import logsubexp, log_integrate_log_trap


def compute_weights(samples, nlive):
    """
    Returns the log-evidence and log-weights for the log-likelihood samples
    assumed to the result of nested sampling with nlive live points

    Parameters
    ----------
    samples : array_like
        Log-likelihood samples.
    nlive : int
        Number of live points used in nested sampling.

    Returns
    -------
    float
        The computed log-evidence.
    array_like
        Array of computed weights (already normalised by the log-evidence).
    """
    samples = np.asarray(samples)
    start_data = np.concatenate(([float("-inf")], samples[:-nlive]))
    end_data = samples[-nlive:]

    log_wts = np.zeros(samples.shape[0])

    log_vols_start = np.cumsum(
        np.ones(len(start_data) + 1) * np.log1p(-1.0 / nlive)
    ) - np.log1p(-1 / nlive)
    log_vols_end = np.zeros(len(end_data))
    log_vols_end[-1] = np.NINF
    log_vols_end[0] = log_vols_start[-1] + np.log1p(-1.0 / nlive)
    for i in range(len(end_data) - 1):
        log_vols_end[i + 1] = log_vols_end[i] + np.log1p(-1.0 / (nlive - i))

    log_likes = np.concatenate((start_data, end_data, [end_data[-1]]))

    log_vols = np.concatenate((log_vols_start, log_vols_end))
    log_ev = log_integrate_log_trap(log_likes, log_vols)

    log_dXs = logsubexp(log_vols[:-1], log_vols[1:])
    log_wts = log_likes[1:-1] + log_dXs[:-1]

    log_wts -= log_ev

    return log_ev, log_wts


def draw_posterior_samples(nested_samples, nlive):
    """
    Draw posterior samples given the nested samples and number of live points.

    Parameters
    ----------
    nested_samples : structured array
        Array of nested samples.
    nlive : int
        Number of live points used during nested sampling.

    Returns
    -------
    array_like
        Samples from the posterior distribution.
    """
    log_Z, log_w = compute_weights(nested_samples["logL"], nlive)
    log_w -= np.max(log_w)
    log_u = np.log(np.random.rand(nested_samples.size))
    return nested_samples[log_w > log_u]



# -*- coding: utf-8 -*-
"""
Tests for the FlowSampler class.
"""
import os
import signal
import sys
import time
from threading import Thread

import pytest
from nessai.flowsampler import FlowSampler
import numpy as np
from unittest.mock import MagicMock, create_autospec, patch


@pytest.fixture()
def flow_sampler():
    return create_autospec(FlowSampler)


@pytest.fixture(autouse=True)
def reset_handlers():
    """Reset signal handling after each test"""
    yield
    # Just in case the tests were ran on Windows
    try:
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
    except AttributeError:
        print("Cannot set signal attributes on this system.")


@pytest.mark.parametrize("resume", [False, True])
def test_init_no_resume_file(flow_sampler, tmp_path, resume):
    """Test the init method when there is no run to resume from"""

    model = MagicMock()
    output = tmp_path / "init"
    output.mkdir()
    output = str(output)
    resume = resume
    exit_code = 131
    max_threads = 2
    resume_file = "test.pkl"
    kwargs = dict(
        nlive=1000,
        pytorch_threads=1,
    )

    flow_sampler.save_kwargs = MagicMock()

    with patch(
        "nessai.flowsampler.NestedSampler", return_value="ns"
    ) as mock, patch("nessai.flowsampler.configure_threads") as mock_threads:
        FlowSampler.__init__(
            flow_sampler,
            model,
            output=output,
            resume=resume,
            exit_code=exit_code,
            max_threads=max_threads,
            resume_file=resume_file,
            **kwargs,
        )

    mock_threads.assert_called_once_with(
        max_threads=max_threads,
        pytorch_threads=1,
        n_pool=None,
    )

    mock.assert_called_once_with(
        model,
        output=os.path.join(output, ""),
        resume_file=resume_file,
        **kwargs,
    )

    assert flow_sampler.ns == "ns"

    flow_sampler.save_kwargs.assert_called_once_with(kwargs)


@pytest.mark.parametrize(
    "test_old, error",
    [(False, None), (True, RuntimeError), (True, FileNotFoundError)],
)
def test_init_resume(flow_sampler, tmp_path, test_old, error):
    """Test the init method when the sampler should resume.

    Tests the case where the first file works and the case where the first
    file fails but the old method works.
    """
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    resume = True
    exit_code = 131
    max_threads = 2
    resume_file = "test.pkl"
    weights_file = "model.pt"
    flow_config = dict(lr=0.1)

    if test_old:
        expected_rf = output / (resume_file + ".old")
        side_effect = [error, "ns"]
    else:
        expected_rf = output / resume_file
        side_effect = ["ns"]
    expected_rf.write_text("contents")

    output = str(output)
    expected_rf = str(expected_rf)

    kwargs = dict(
        nlive=1000,
        pytorch_threads=1,
        flow_config=flow_config,
    )

    flow_sampler.save_kwargs = MagicMock()

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ) as mock_resume, patch(
        "nessai.flowsampler.configure_threads"
    ) as mock_threads:
        FlowSampler.__init__(
            flow_sampler,
            model,
            output=output,
            resume=resume,
            exit_code=exit_code,
            max_threads=max_threads,
            resume_file=resume_file,
            weights_file=weights_file,
            **kwargs,
        )

    mock_threads.assert_called_once_with(
        max_threads=max_threads,
        pytorch_threads=1,
        n_pool=None,
    )

    mock_resume.assert_called_with(
        expected_rf,
        model,
        flow_config=flow_config,
        weights_file=weights_file,
    )

    assert flow_sampler.ns == "ns"

    flow_sampler.save_kwargs.assert_called_once_with(kwargs)


def test_init_resume_error_cannot_resume(flow_sampler, tmp_path):
    """Assert an error is raised if neither file loads"""
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    resume = True
    resume_file = "test.pkl"
    expected_rf = output / (resume_file + ".old")
    expected_rf.write_text("contents")
    side_effect = [RuntimeError, RuntimeError]
    output = str(output)
    expected_rf = str(expected_rf)

    assert os.path.exists(expected_rf)

    with patch(
        "nessai.flowsampler.NestedSampler.resume", side_effect=side_effect
    ), patch("nessai.flowsampler.configure_threads"), pytest.raises(
        RuntimeError
    ) as excinfo:
        FlowSampler.__init__(
            flow_sampler,
            model,
            output=output,
            resume=resume,
            resume_file=resume_file,
            flow_config=None,
        )
    assert "Could not resume sampler with error: " in str(excinfo.value)


def test_init_resume_error_no_file(flow_sampler, tmp_path):
    """Assert an error is raised if resume=True and a resume_file is \
        not specified.
    """
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("nessai.flowsampler.configure_threads"), pytest.raises(
        RuntimeError
    ) as excinfo:
        FlowSampler.__init__(
            flow_sampler,
            model,
            output=output,
            resume=True,
            resume_file=None,
        )
    assert "`resume_file` must be specified" in str(excinfo.value)


def test_init_signal_handling_enabled(flow_sampler, tmp_path):
    """Assert signal.signal is called when signal handling is enabled."""
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal") as mocked_fn:
        FlowSampler.__init__(
            flow_sampler, model, output=output, signal_handling=True
        )
    mocked_fn.assert_called()


def test_init_signal_handling_disabled(flow_sampler, tmp_path):
    """Assert signal handling is not configure when disabled"""
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal") as mocked_fn:
        FlowSampler.__init__(
            flow_sampler, model, output=output, signal_handling=False
        )
    mocked_fn.assert_not_called()


def test_init_signal_handling_error(flow_sampler, tmp_path, caplog):
    """Assert signal handling is skipped if an error is raised."""
    model = MagicMock()
    output = tmp_path / "test"
    output.mkdir()
    output = str(output)

    with patch("signal.signal", side_effect=AttributeError):
        FlowSampler.__init__(
            flow_sampler, model, output=output, signal_handling=True
        )
    assert "Cannot set signal attributes" in str(caplog.text)


@pytest.mark.parametrize("save", [False, True])
@pytest.mark.parametrize("plot", [False, True])
@patch(
    "nessai.flowsampler.draw_posterior_samples", return_value=np.array([0.1])
)
@patch("nessai.plot.plot_live_points")
@patch("nessai.plot.plot_indices")
def test_run(
    mock_plot_indices, mock_plot_post, mock_draw_post, flow_sampler, save, plot
):
    """Test the run method"""
    nlive = 10
    log_Z = -5.0
    nested_samples = [0.1, 1.0, 10.0]
    insertion_indices = [1, 2, 3]
    output = os.getcwd()
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.nlive = nlive
    flow_sampler.ns.insertion_indices = insertion_indices
    flow_sampler.output = output
    flow_sampler.ns.initialise = MagicMock()
    flow_sampler.ns.nested_sampling_loop = MagicMock(
        return_value=[log_Z, nested_samples]
    )
    flow_sampler.ns.state = MagicMock()
    flow_sampler.ns.state.plot_state = MagicMock()
    flow_sampler.save_results = MagicMock()

    FlowSampler.run(flow_sampler, save=save, plot=plot)

    flow_sampler.ns.initialise.assert_called_once()
    mock_draw_post.assert_called_once_with(nested_samples, nlive)
    if save:
        flow_sampler.save_results.assert_called_once()
    else:
        flow_sampler.save_results.assert_not_called()

    if plot:
        mock_plot_indices.assert_called_once_with(
            insertion_indices,
            nlive,
            filename=os.path.join(output, "insertion_indices.png"),
        )
        mock_plot_post.assert_called_once_with(
            flow_sampler.posterior_samples,
            filename=os.path.join(output, "posterior_distribution.png"),
        )
        flow_sampler.ns.state.plot.assert_called_once_with(
            os.path.join(output, "logXlogL.png"),
        )
    else:
        mock_plot_indices.assert_not_called()
        mock_plot_post.assert_not_called()
        flow_sampler.ns.state.plot.assert_not_called()

    assert flow_sampler.logZ == log_Z
    assert flow_sampler.nested_samples == nested_samples
    np.testing.assert_array_equal(
        flow_sampler.posterior_samples, np.array([0.1])
    )


@pytest.mark.parametrize("test_class", [False, True])
def test_save_kwargs(flow_sampler, tmpdir, test_class):
    """Test the save kwargs method.

    If `test_class` is true, tests the case of the flow class being a class
    rather than a string.
    """
    kwargs = dict(nlive=10, a=np.array([0.1]))
    if test_class:
        from nessai.proposal import FlowProposal

        kwargs["flow_class"] = FlowProposal
    else:
        kwargs["flow_class"] = "flowproposal"

    flow_sampler.output = str(tmpdir.mkdir("test"))

    FlowSampler.save_kwargs(flow_sampler, kwargs)

    assert os.path.exists(os.path.join(flow_sampler.output, "config.json"))


def test_save_results(flow_sampler, tmpdir):
    """Test the save results method"""
    from datetime import timedelta

    output = str(tmpdir.mkdir("test"))
    filename = os.path.join(output, "result.json")

    ns = MagicMock()
    ns.nlive = 1
    ns.iteration = 3
    ns.min_likelihood = [-3, -2, 1]
    ns.max_likelihood = [1, 2, 3]
    ns.likelihood_evaluations = 3
    ns.logZ_history = [1, 2, 3]
    ns.mean_acceptance_history = [1, 2, 3]
    ns.rolling_p = [0.5]
    ns.population_iterations = []
    ns.population_acceptance = []
    ns.training_iterations = []
    ns.insertion_indices = []
    ns.sampling_time = timedelta()
    ns.training_time = timedelta()
    ns.proposal_population_time = timedelta()
    ns.likelihood_evaluation_time = timedelta(2)

    flow_sampler.ns = ns
    flow_sampler.posterior_samples = np.array([0.1], dtype=[("x", "f8")])
    flow_sampler.nested_samples = np.array([0.1], dtype=[("x", "f8")])

    FlowSampler.save_results(flow_sampler, filename)

    assert os.path.exists(filename)


def test_safe_exit(flow_sampler):
    """Test the safe exit method."""
    flow_sampler.exit_code = 130
    flow_sampler.ns = MagicMock()
    flow_sampler.ns.checkpoint = MagicMock()
    flow_sampler.ns.model = MagicMock()
    flow_sampler.ns.model.close_pool = MagicMock()

    with patch("sys.exit") as mock_exit:
        FlowSampler.safe_exit(flow_sampler, signum=2)

    mock_exit.assert_called_once_with(130)
    flow_sampler.ns.checkpoint.assert_called_once()
    flow_sampler.ns.model.close_pool.assert_called_once_with(code=2)


@pytest.mark.parametrize(
    "kwargs", [dict(n_pool=None), dict(max_threads=3, n_pool=2)]
)
@pytest.mark.integration_test
@pytest.mark.timeout(30)
@pytest.mark.skip_on_windows
def test_signal_handling(tmp_path, caplog, model, kwargs, mp_context):
    """Test the signal handling in nessai.

    Test is based on a similar test in bilby which is in turn based on: \
        https://stackoverflow.com/a/49615525/18400311
    """
    output = tmp_path / "output"
    output.mkdir()

    with patch("multiprocessing.Pool", mp_context.Pool):
        fs = FlowSampler(
            model,
            output=output,
            nlive=500,
            poolsize=1000,
            exit_code=5,
            signal_handling=True,
            **kwargs,
        )

    pid = os.getpid()

    def trigger_signal():
        time.sleep(4)
        os.kill(pid, signal.SIGINT)

    thread = Thread(target=trigger_signal)
    thread.daemon = True
    thread.start()

    with pytest.raises(SystemExit):
        try:
            while True:
                fs.run(save=False, plot=False)
        except SystemExit as error:
            assert error.code == 5
            raise

    assert f"Trying to safely exit with code {signal.SIGINT}" in str(
        caplog.text
    )


@pytest.mark.integration_test
@pytest.mark.timeout(30)
@pytest.mark.skip_on_windows
def test_signal_handling_disabled(tmp_path, caplog, model):
    """Assert signal handling is correctly disabled.

    Test is based on a similar test in bilby which is in turn based on: \
        https://stackoverflow.com/a/49615525/18400311
    """
    output = tmp_path / "output"
    output.mkdir()

    fs = FlowSampler(
        model,
        output=output,
        nlive=500,
        poolsize=1000,
        exit_code=4,
        signal_handling=False,
    )

    pid = os.getpid()

    def trigger_signal():
        time.sleep(4)
        os.kill(pid, signal.SIGINT)

    thread = Thread(target=trigger_signal)
    thread.daemon = True
    thread.start()

    # Need to catch SIGINT to prevent tests for exiting
    def handler(signum=None, frame=None):
        sys.exit(10)

    signal.signal(signal.SIGINT, handler)

    with pytest.raises(SystemExit):
        try:
            while True:
                fs.run(save=False, plot=False)
        except SystemExit as error:
            assert error.code == 10
            raise

    assert "Signal handling is disabled" in str(caplog.text)



# -*- coding: utf-8 -*-
"""Test methods related to reparameterisations"""
import numpy as np
from nessai.livepoint import get_dtype, numpy_array_to_live_points
from nessai.proposal import FlowProposal
from nessai.reparameterisations import (
    NullReparameterisation,
    RescaleToBounds,
    get_reparameterisation,
)
from nessai.utils.testing import assert_structured_arrays_equal
import pytest
from unittest.mock import MagicMock, Mock, call, patch


@pytest.fixture
def proposal(proposal):
    """Specific mocked proposal for reparameterisation tests"""
    proposal.use_default_reparameterisations = False
    return proposal


@pytest.fixture
def dummy_rc():
    """Dummy reparameteristation class"""
    m = MagicMock()
    m.__name__ = "DummyReparameterisation"
    return m


@pytest.fixture
def dummy_cmb_rc():
    """Dummy combined reparameteristation class"""
    m = MagicMock()
    m.add_reparameterisation = MagicMock()
    return m


def test_default_reparameterisation(proposal):
    """Test to make sure default reparameterisation does not cause errors
    for default proposal.
    """
    FlowProposal.add_default_reparameterisations(proposal)


@patch("nessai.reparameterisations.get_reparameterisation")
def test_get_reparamaterisation(mocked_fn, proposal):
    """Make sure the underlying function is called"""
    FlowProposal.get_reparameterisation(proposal, "angle")
    assert mocked_fn.called_once_with("angle")


def test_configure_reparameterisations_dict(proposal, dummy_cmb_rc, dummy_rc):
    """Test configuration for reparameterisations dictionary.

    Also tests to make sure boundary inversion is set.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x"]
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(dummy_rc, {"boundary_inversion": True})
    )
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x"]

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal, {"x": {"reparameterisation": "default"}}
        )

    proposal.get_reparameterisation.assert_called_once_with("default")
    proposal.add_default_reparameterisations.assert_called_once()
    dummy_rc.assert_called_once_with(
        prior_bounds={"x": [-1, 1]}, parameters="x", boundary_inversion=True
    )
    mocked_class.assert_called_once()
    # fmt: off
    proposal._reparameterisation.add_reparameterisations \
        .assert_called_once_with("r")
    # fmt: on

    assert proposal.boundary_inversion is True
    assert proposal.names == ["x"]


@patch("nessai.proposal.flowproposal.CombinedReparameterisation")
def test_configure_reparameterisations_dict_w_params(
    mocked_class, proposal, dummy_rc, dummy_cmb_rc
):
    """Test configuration for reparameterisations dictionary with parameters.

    For example:

        {'x': {'reparmeterisation': 'default', 'parameters': 'y'}}

    This should add both x and y to the reparameterisation.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x", "y"]
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal,
            {"x": {"reparameterisation": "default", "parameters": ["y"]}},
        )

    proposal.get_reparameterisation.assert_called_once_with("default")
    proposal.add_default_reparameterisations.assert_called_once()
    dummy_rc.assert_called_once_with(
        prior_bounds={"x": [-1, 1], "y": [-1, 1]},
        parameters=["y", "x"],
    )
    mocked_class.assert_called_once()
    # fmt: off
    proposal._reparameterisation.add_reparameterisations \
        .assert_called_once_with("r")
    # fmt: on

    assert proposal.names == ["x", "y"]


def test_configure_reparameterisations_requires_prime_prior(
    proposal, dummy_rc, dummy_cmb_rc
):
    """
    Test configuration that requires a prime prior but the prime prior is
    missing.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x", "y"]
    dummy_cmb_rc.has_prime_prior = False
    dummy_cmb_rc.requires_prime_prior = True
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ), pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal,
            {"x": {"reparameterisation": "default", "parameters": ["y"]}},
        )

    assert "One or more reparameterisations require " in str(excinfo.value)


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_dict_missing(mocked_class, proposal):
    """
    Test configuration for reparameterisations dictionary when missing
    the reparameterisation for a parameter.

    This should raise a runtime error.
    """
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {"x": {"scale": 1.0}}
        )

    assert "No reparameterisation found for x" in str(excinfo.value)


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_str(mocked_class, proposal):
    """Test configuration for reparameterisations dictionary from a str"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(proposal, {"x": "default"})

    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ["x_prime", "y"]
    assert proposal.rescale_parameters == ["x"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x_prime", "y"]
    assert mocked_class.called_once


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_dict_reparam(mocked_class, proposal):
    """Test configuration for reparameterisations dictionary"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(
        proposal, {"default": {"parameters": ["x"]}}
    )

    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ["x_prime", "y"]
    assert proposal.rescale_parameters == ["x"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x_prime", "y"]
    assert mocked_class.called_once


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_none(mocked_class, proposal):
    """Test configuration when input is None"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(proposal, None)
    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ["x", "y"]

    assert proposal.rescale_parameters == []
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x", "y"]
    assert all(
        [
            isinstance(r, NullReparameterisation)
            for r in proposal._reparameterisation.reparameterisations.values()
        ]
    )
    assert mocked_class.called_once


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_fallback(mocked_class, proposal):
    """Test configuration when input is None"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.names = ["x", "y"]
    proposal.fallback_reparameterisation = "default"
    FlowProposal.configure_reparameterisations(proposal, None)
    proposal.add_default_reparameterisations.assert_called_once()
    assert proposal.rescaled_names == ["x_prime", "y_prime"]

    assert proposal.rescale_parameters == ["x", "y"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == [
        "x_prime",
        "y_prime",
    ]
    assert all(
        [
            isinstance(r, RescaleToBounds)
            for r in proposal._reparameterisation.reparameterisations.values()
        ]
    )
    assert mocked_class.called_once


def test_configure_reparameterisations_incorrect_type(proposal):
    """Assert an error is raised when input is not a dictionary"""
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, ["default"])
    assert "must be a dictionary" in str(excinfo.value)


def test_configure_reparameterisations_incorrect_config_type(proposal):
    """Assert an error is raised when the config for a key is not a dictionary
    or a known reparameterisation.
    """
    proposal.names = ["x"]
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, {"x": ["a"]})
    assert "Unknown config type" in str(excinfo.value)


@pytest.mark.parametrize(
    "reparam",
    [{"z": {"reparameterisation": "sine"}}, {"sine": {"parameters": ["x"]}}],
)
def test_configure_reparameterisation_unknown(proposal, reparam):
    """
    Assert an error is raised if an unknown reparameterisation or parameters
    is passed.
    """
    proposal.names = ["x"]
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, reparam)
    assert "is not a parameter in the model or a known" in str(excinfo.value)


def test_configure_reparameterisation_no_parameters(proposal, dummy_rc):
    """Assert an error is raised if no parameters are specified"""
    proposal.names = ["x"]
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {"default": {"update_bounds": True}}
        )
    assert "No parameters key" in str(excinfo.value)


def test_set_rescaling_with_model(proposal, model):
    """
    Test setting the rescaling when the model contains reparmaeterisations.
    """
    proposal.model = model
    proposal.model.reparameterisations = {"x": "default"}
    proposal.set_boundary_inversion = MagicMock()

    def update(self):
        proposal.rescale_parameters = ["x"]
        proposal.rescaled_names = ["x_prime"]

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.set_boundary_inversion.assert_called_once()
    proposal.configure_reparameterisations.assert_called_once_with(
        {"x": "default"}
    )
    assert proposal.reparameterisations == {"x": "default"}
    assert proposal.rescaled_names == ["x_prime"]


def test_set_rescaling_with_reparameterisations(proposal, model):
    """
    Test setting the rescaling when a reparameterisations dict is defined.
    """
    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = {"x": "default"}
    proposal.set_boundary_inversion = MagicMock()

    def update(self):
        proposal.rescale_parameters = ["x"]
        proposal.rescaled_names = ["x_prime"]

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.set_boundary_inversion.assert_called_once()
    proposal.configure_reparameterisations.assert_called_once_with(
        {"x": "default"}
    )
    assert proposal.reparameterisations == {"x": "default"}
    assert proposal.rescaled_names == ["x_prime"]


@pytest.mark.parametrize("update_bounds", [True, False])
@pytest.mark.parametrize(
    "rescale_parameters",
    [
        (False, ["x", "y"]),
        (True, ["x_prime", "y_prime"]),
        (["x"], ["x_prime", "y"]),
    ],
)
def test_set_rescaling_parameters_list(
    proposal, model, update_bounds, rescale_parameters
):
    """Test setting rescaling without reparameterisations."""
    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = None
    proposal.configure_reparameterisations = MagicMock()

    proposal.rescale_parameters = rescale_parameters[0]
    proposal.rescale_bounds = [-1, 1]
    proposal.update_bounds = update_bounds

    FlowProposal.set_rescaling(proposal)

    proposal.configure_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == rescale_parameters[1]

    if rescale_parameters[0]:
        assert proposal._rescale_factor == 2.0
        assert proposal._rescale_shift == -1.0
        assert proposal._min == {"x": -5, "y": -5}
        assert proposal._max == {"x": 5, "y": 5}


def test_default_rescale(proposal):
    """Test the default rescaling method."""
    x = np.array([[1, 2], [3, 4]])
    x_prime, log_j = FlowProposal.rescale(proposal, x)
    np.testing.assert_array_equal(x, x_prime)
    assert (log_j == 0.0).all()


def test_default_inverse_rescale(proposal):
    """Test the default inverse rescaling method."""
    x_prime = np.array([[1, 2], [3, 4]])
    x, log_j = FlowProposal.inverse_rescale(proposal, x_prime)
    np.testing.assert_array_equal(x, x_prime)
    assert (log_j == 0.0).all()


@pytest.mark.parametrize("n", [1, 10])
def test_rescale_w_reparameterisation(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"])
    x["logL"] = np.random.randn(n)
    x["logP"] = np.random.randn(n)
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ["x_prime", "y_prime"]
    )
    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime"])
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.reparameterise = MagicMock(
        return_value=[x, x_prime, np.ones(x.size)]
    )

    x_prime_out, log_j = FlowProposal._rescale_w_reparameterisation(
        proposal, x, compute_radius=False, test="lower"
    )

    np.testing.assert_array_equal(
        x_prime[["x_prime", "y_prime"]], x_prime_out[["x_prime", "y_prime"]]
    )
    np.testing.assert_array_equal(
        x[["logP", "logL"]], x_prime_out[["logP", "logL"]]
    )
    proposal._reparameterisation.reparameterise.assert_called_once()


@pytest.mark.parametrize("n", [1, 10])
def test_inverse_rescale_w_reparameterisation(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"]).squeeze()
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ["x_prime", "y_prime"]
    )
    x_prime["logL"] = np.random.randn(n)
    x_prime["logP"] = np.random.randn(n)
    proposal.x_dtype = get_dtype(["x", "y"])
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.inverse_reparameterise = MagicMock(
        return_value=[x, x_prime, np.ones(x.size)]
    )

    x_out, log_j = FlowProposal._inverse_rescale_w_reparameterisation(
        proposal, x_prime
    )

    np.testing.assert_array_equal(x[["x", "y"]], x_out[["x", "y"]])
    np.testing.assert_array_equal(
        x_prime[["logP", "logL"]], x_out[["logP", "logL"]]
    )
    proposal._reparameterisation.inverse_reparameterise.assert_called_once()


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("compute_radius", [False, True])
def test_rescale_to_bounds(proposal, model, n, compute_radius):
    """Test the default rescaling to bounds.

    Also tests the log Jacobian determinant.
    """
    model.names.append("z")
    model.bounds["z"] = [-10, 10]
    x = numpy_array_to_live_points(
        np.random.randn(n, 3), ["x", "y", "z"]
    ).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 3]), ["x_prime", "y_prime", "z"]
    )

    x_prime_expected["x_prime"] = 2 * (x["x"] + 5) / 10 - 1
    x_prime_expected["y_prime"] = 2 * (x["y"] + 4) / 8 - 1
    x_prime_expected["z"] = x["z"]

    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime", "z"])

    proposal.names = ["x", "y", "z"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime", "z"]
    proposal.model = model
    proposal.boundary_inversion = []

    proposal._rescale_factor = 2.0
    proposal._rescale_shift = -1.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}

    x_prime, log_j = FlowProposal._rescale_to_bounds(
        proposal, x, compute_radius=compute_radius
    )

    np.testing.assert_allclose(log_j, -np.log(20), rtol=1e-14)
    assert_structured_arrays_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("inputs", [("split", True), ("duplicate", False)])
@pytest.mark.parametrize("bound", ["lower", "upper"])
def test_rescale_to_bounds_w_inversion_duplicate(
    proposal, model, n, inputs, bound
):
    """Test the default rescaling with inversion using duplication."""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"]).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ["x_prime", "y_prime"]
    )

    x_prime_expected["x_prime"] = (x["x"] + 5) / 10
    x_prime_expected["y_prime"] = (x["y"] + 4) / 8

    if bound == "lower":
        x_prime_expected = np.concatenate([x_prime_expected, x_prime_expected])
        x_prime_expected["x_prime"][n:] *= -1
    elif bound == "upper":
        x_prime_expected["x_prime"] = 1 - x_prime_expected["x_prime"]
        x_prime_expected = np.concatenate([x_prime_expected, x_prime_expected])
        x_prime_expected["x_prime"][n:] *= -1

    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime"])

    proposal.names = ["x", "y"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime"]
    proposal.model = model
    proposal.boundary_inversion = ["x"]

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}
    proposal._edges = {"x": None}
    proposal.detect_edges_kwargs = {"k": 2}
    proposal.inversion_type = inputs[0]

    with patch(
        "nessai.proposal.flowproposal.detect_edge", return_value=bound
    ) as mock_detect_edge:
        x_prime, log_j = FlowProposal._rescale_to_bounds(
            proposal, x, compute_radius=inputs[1], test=True
        )

    np.testing.assert_array_equal(
        mock_detect_edge.call_args[0][0], x_prime_expected["x_prime"][:n]
    )
    assert mock_detect_edge.call_args[1]["test"] is True
    assert mock_detect_edge.call_args[1]["k"] == 2

    np.testing.assert_equal(log_j, -np.log(80))
    assert_structured_arrays_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("bound", ["lower", "upper"])
def test_rescale_to_bounds_w_inversion_split(proposal, model, n, bound):
    """Test the default rescaling with inversion using splitting."""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"]).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ["x_prime", "y_prime"]
    )

    x_prime_expected["x_prime"] = (x["x"] + 5) / 10
    x_prime_expected["y_prime"] = (x["y"] + 4) / 8

    inv = np.random.choice(n, n // 2, replace=False)

    if bound == "lower":
        x_prime_expected["x_prime"][inv] *= -1
    elif bound == "upper":
        x_prime_expected["x_prime"] = 1 - x_prime_expected["x_prime"]
        x_prime_expected["x_prime"][inv] *= -1

    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime"])

    proposal.names = ["x", "y"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime"]
    proposal.model = model
    proposal.boundary_inversion = ["x"]

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}
    proposal._edges = {"x": None}
    proposal.detect_edges_kwargs = {"k": 2}
    proposal.inversion_type = "split"

    with patch(
        "nessai.proposal.flowproposal.detect_edge", return_value=bound
    ) as mock_detect_edge, patch("numpy.random.choice", return_value=inv):
        x_prime, log_j = FlowProposal._rescale_to_bounds(
            proposal, x, compute_radius=False, test=True
        )

    np.testing.assert_array_equal(
        mock_detect_edge.call_args[0][0], x_prime_expected["x_prime"][:n]
    )
    assert mock_detect_edge.call_args[1]["test"] is True
    assert mock_detect_edge.call_args[1]["k"] == 2

    np.testing.assert_equal(log_j, -np.log(80))
    assert_structured_arrays_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("bound", ["lower", "upper"])
def test_rescale_to_bounds_w_inversion_false(proposal, model, n, bound):
    """
    Test the default rescaling without inversion and with edges already set.
    """
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"]).squeeze()
    x_prime_expected = numpy_array_to_live_points(
        np.zeros([n, 2]), ["x_prime", "y_prime"]
    )

    x_prime_expected["x_prime"] = (x["x"] + 5) / 10
    x_prime_expected["y_prime"] = (x["y"] + 4) / 8

    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime"])

    proposal.names = ["x", "y"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime"]
    proposal.model = model
    proposal.boundary_inversion = ["x"]

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}
    proposal._edges = {"x": False}
    proposal.detect_edges_kwargs = {"k": 2}
    proposal.inversion_type = "split"

    with patch(
        "nessai.proposal.flowproposal.detect_edge", return_value=bound
    ) as mock_detect_edge:
        x_prime, log_j = FlowProposal._rescale_to_bounds(
            proposal, x, compute_radius=False, test=True
        )

    mock_detect_edge.assert_not_called()

    np.testing.assert_equal(log_j, -np.log(80))
    assert_structured_arrays_equal(x_prime, x_prime_expected)


@pytest.mark.parametrize("n", [1, 10])
def test_inverse_rescale_to_bounds(proposal, model, n):
    """Test the default method for the inverse rescaling.

    Also tests the log Jacobian determinant.
    """
    model.names.append("z")
    model.bounds["z"] = [-10, 10]
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 3), ["x_prime", "y_prime", "z"]
    ).squeeze()
    x_expected = numpy_array_to_live_points(np.zeros([n, 3]), ["x", "y", "z"])

    x_expected["x"] = 10.0 * (x_prime["x_prime"] + 1.0) / 2.0 - 5.0
    x_expected["y"] = 8.0 * (x_prime["y_prime"] + 1.0) / 2.0 - 4.0
    x_expected["z"] = x_prime["z"]

    proposal.x_dtype = get_dtype(["x", "y", "z"])

    proposal.names = ["x", "y", "z"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime", "z"]
    proposal.model = model
    proposal.boundary_inversion = []

    proposal._rescale_factor = 2.0
    proposal._rescale_shift = -1.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}

    x, log_j = FlowProposal._inverse_rescale_to_bounds(proposal, x_prime)

    np.testing.assert_allclose(log_j, np.log(20), rtol=1e-14)
    assert_structured_arrays_equal(x, x_expected)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("itype", ["lower", "upper"])
def test_inverse_rescale_to_bounds_w_inversion(proposal, model, n, itype):
    """Test the default method for the inverse rescaling with inversion.

    Also tests the log Jacobian determinant.
    """
    model.names.append("z")
    model.bounds["z"] = [-10, 10]
    x_prime = numpy_array_to_live_points(
        np.random.uniform(-1, 1, (n, 3)), ["x_prime", "y_prime", "z"]
    ).squeeze()

    x_expected = numpy_array_to_live_points(np.zeros([n, 3]), ["x", "y", "z"])

    if itype == "lower":
        x_expected["x"] = 10.0 * np.abs(x_prime["x_prime"]) - 5.0
    elif itype == "upper":
        x_expected["x"] = 10.0 * (1 - np.abs(x_prime["x_prime"])) - 5.0

    x_expected["y"] = 8.0 * x_prime["y_prime"] - 4.0
    x_expected["z"] = x_prime["z"]

    proposal.x_dtype = get_dtype(["x", "y", "z"])

    proposal.names = ["x", "y", "z"]
    proposal.rescale_parameters = ["x", "y"]
    proposal.rescaled_names = ["x_prime", "y_prime", "z"]
    proposal.model = model
    proposal.boundary_inversion = ["x"]
    proposal.inversion_type = "split"
    proposal._edges = {"x": itype}

    proposal._rescale_factor = 1.0
    proposal._rescale_shift = 0.0
    proposal._min = {"x": -5, "y": -4}
    proposal._max = {"x": 5, "y": 4}

    x, log_j = FlowProposal._inverse_rescale_to_bounds(proposal, x_prime)

    np.testing.assert_equal(log_j, np.log(80))
    assert_structured_arrays_equal(x, x_expected)


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling(proposal, has_inversion):
    """Test the method that tests the rescaling at runtime"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.check_state = MagicMock()

    FlowProposal.verify_rescaling(proposal)

    proposal.check_state.assert_has_calls(4 * [call(x)])
    # Should call 4 different test cases
    calls = [
        call(x, test="lower"),
        call(x, test="upper"),
        call(x, test=False),
        call(x, test=None),
    ]
    proposal.rescale.assert_has_calls(calls)
    proposal.inverse_rescale.assert_has_calls(4 * [call(x_prime)])


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling_invertible_error(proposal, has_inversion):
    """Assert an error is raised if the rescaling is not invertible"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()[::-1]
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Rescaling is not invertible for x" in str(excinfo.value)


def test_verify_rescaling_duplicate_error(proposal):
    """Assert an error is raised if the duplication is missing samples"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2, -2, -2])
    x_out = np.array([[1], [3], [4], [5]], dtype=[("x", "f8")])
    log_j_inv = np.array([2, 2, 2, 2])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Duplicate samples must map to same input" in str(excinfo.value)


def test_verify_rescaling_duplicate_error_nans(proposal):
    """Assert an error is raised if the duplication is missing samples"""
    x = np.array([[np.nan], [np.nan]], dtype=[("x", "f8")])
    x_prime = np.array([1.0, 2.0])
    log_j = np.array([-2, -2, -2, -2])
    x_out = np.array([[np.nan], [np.nan], [4], [np.nan]], dtype=[("x", "f8")])
    log_j_inv = np.array([2, 2, 2, 2])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Rescaling is not invertible for x (NaNs)" in str(excinfo.value)


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling_jacobian_error(proposal, has_inversion):
    """Assert an error is raised if the Jacobian is not invertible"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 1])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model = MagicMock()
    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Rescaling Jacobian is not invertible" in str(excinfo.value)


def test_check_state_boundary_inversion_default(proposal):
    """Test the check state method for boundary inversion"""
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal._reparameterisation = None
    proposal._edges = {"x": "lower", "y": "upper"}
    proposal.boundary_inversion = ["x", "y"]
    proposal.update_bounds = False
    FlowProposal.check_state(proposal, x)
    assert all(v is None for v in proposal._edges.values())


def test_check_state_boundary_inversion_reparameterisations(proposal):
    """
    Test the check state method for boundary inversion with
    reparameterisations.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal._reparameterisation = Mock()
    proposal._reparameterisation.reset_inversion = MagicMock()
    proposal.boundary_inversion = ["x", "y"]
    proposal.update_bounds = False
    FlowProposal.check_state(proposal, x)
    proposal._reparameterisation.reset_inversion.assert_called_once()


def test_check_state_update_bounds_default(proposal, model):
    """
    Test the check state method for updating bounds.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal.model = model
    excepted_min = (np.min(x["x"]), np.min(x["y"]))
    excepted_max = (np.max(x["x"]), np.max(x["y"]))
    proposal._reparameterisation = None
    proposal.boundary_inversion = False
    proposal.update_bounds = True
    proposal._min = {"x": -np.inf, "y": -np.inf}
    proposal._max = {"x": np.inf, "y": np.inf}
    FlowProposal.check_state(proposal, x)

    assert proposal._min["x"] == excepted_min[0]
    assert proposal._min["y"] == excepted_min[1]
    assert proposal._max["x"] == excepted_max[0]
    assert proposal._max["y"] == excepted_max[1]


def test_check_state_update_bounds_reparameterisations(proposal):
    """
    Test the check state method for updating bounds.
    """
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal._reparameterisation = Mock()
    proposal._reparameterisation.update_bounds = MagicMock()
    proposal.boundary_inversion = False
    proposal.update_bounds = True

    FlowProposal.check_state(proposal, x)
    proposal._reparameterisation.update_bounds.assert_called_once()



# -*- coding: utf-8 -*-
"""
Main code that handles running and checkpoiting the sampler.
"""
import json
import logging
import os
import signal
import sys

import numpy as np

from . import __version__ as version
from .livepoint import live_points_to_dict
from .nestedsampler import NestedSampler
from .posterior import draw_posterior_samples
from .utils import NessaiJSONEncoder, configure_threads


logger = logging.getLogger(__name__)


class FlowSampler:
    """
    Main class to handle running the nested sampler.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model.
    output : str, optional
        Output directory
    resume : bool, optional
        If True try to resume the sampler is the resume file exists.
    resume_file : str, optional
        File to resume sampler from.
    weights_files : str, optional
        Weights files used to resume sampler that replaces the weights file
        saved internally.
    max_threads : int, optional
        Maximum number of threads to use. If ``None`` torch uses all available
        threads.
    signal_handling : bool
        Enable or disable signal handling.
    exit_code : int, optional
        Exit code to use when forceably exiting the sampler.
    kwargs :
        Keyword arguments passed to :obj:`~nessai.nestedsampler.NestedSampler`.
    """

    def __init__(
        self,
        model,
        output="./",
        resume=True,
        resume_file="nested_sampler_resume.pkl",
        weights_file=None,
        signal_handling=True,
        exit_code=130,
        max_threads=1,
        **kwargs,
    ):

        configure_threads(
            max_threads=max_threads,
            pytorch_threads=kwargs.get("pytorch_threads", None),
            n_pool=kwargs.get("n_pool", None),
        )

        self.exit_code = exit_code

        self.output = os.path.join(output, "")
        if resume:
            if not resume_file:
                raise RuntimeError(
                    "`resume_file` must be specified if resume=True. "
                    f"Current value: {resume_file}"
                )
            if not any(
                (
                    os.path.exists(os.path.join(self.output, f))
                    for f in [resume_file, resume_file + ".old"]
                )
            ):
                logger.warning("No files to resume from, starting sampling")
                self.ns = NestedSampler(
                    model,
                    output=self.output,
                    resume_file=resume_file,
                    **kwargs,
                )
            else:
                try:
                    self.ns = NestedSampler.resume(
                        os.path.join(self.output, resume_file),
                        model,
                        flow_config=kwargs.get("flow_config"),
                        weights_file=weights_file,
                    )
                except (FileNotFoundError, RuntimeError) as e:
                    logger.error(
                        f"Could not load resume file from: {self.output} "
                        f"with error {e}"
                    )
                    try:
                        resume_file += ".old"
                        self.ns = NestedSampler.resume(
                            os.path.join(self.output, resume_file),
                            model,
                            flow_config=kwargs.get("flow_config"),
                            weights_file=weights_file,
                        )
                    except RuntimeError as e:
                        logger.error(
                            "Could not load old resume "
                            f"file from: {self.output}"
                        )
                        raise RuntimeError(
                            "Could not resume sampler " f"with error: {e}"
                        )
        else:
            self.ns = NestedSampler(
                model, output=self.output, resume_file=resume_file, **kwargs
            )

        self.save_kwargs(kwargs)

        if signal_handling:
            try:
                signal.signal(signal.SIGTERM, self.safe_exit)
                signal.signal(signal.SIGINT, self.safe_exit)
                signal.signal(signal.SIGALRM, self.safe_exit)
            except AttributeError:
                logger.critical("Cannot set signal attributes on this system")
        else:
            logger.warning(
                "Signal handling is disabled. nessai will not automatically "
                "checkpoint when exitted."
            )

    def run(self, plot=True, save=True):
        """
        Run the nested samper

        Parameters
        ----------
        plot : bool, optional
            Toggle plots produced once the sampler has converged
        save : bool, optional
            Toggle automatic saving of results
        """
        self.ns.initialise()
        self.logZ, self.nested_samples = self.ns.nested_sampling_loop()
        logger.info((f"Total sampling time: {self.ns.sampling_time}"))

        logger.info("Starting post processing")
        logger.info("Computing posterior samples")
        self.posterior_samples = draw_posterior_samples(
            self.nested_samples, self.ns.nlive
        )
        logger.info(
            f"Returned {self.posterior_samples.size} " "posterior samples"
        )

        if save:
            self.save_results(os.path.join(self.output, "result.json"))

        if plot:
            from nessai import plot

            plot.plot_live_points(
                self.posterior_samples,
                filename=os.path.join(
                    self.output, "posterior_distribution.png"
                ),
            )

            plot.plot_indices(
                self.ns.insertion_indices,
                self.ns.nlive,
                filename=os.path.join(self.output, "insertion_indices.png"),
            )

            self.ns.state.plot(os.path.join(self.output, "logXlogL.png"))

    def save_kwargs(self, kwargs):
        """
        Save the dictionary of keyword arguments used.

        Uses an encoder class to handle numpy arrays.

        Parameters
        ----------
        kwargs : dict
            Dictionary of kwargs to save.
        """
        d = kwargs.copy()
        with open(os.path.join(self.output, "config.json"), "w") as wf:
            json.dump(d, wf, indent=4, cls=NessaiJSONEncoder)

    def save_results(self, filename):
        """
        Save the results from sampling to a specific JSON file.

        Parameters
        ----------
        filename : str
            Name of file to save results to.
        """
        iterations = np.arange(len(self.ns.min_likelihood)) * (
            self.ns.nlive // 10
        )
        iterations[-1] = self.ns.iteration
        d = dict()
        d["version"] = version
        d["history"] = dict(
            iterations=iterations,
            min_likelihood=self.ns.min_likelihood,
            max_likelihood=self.ns.max_likelihood,
            likelihood_evaluations=self.ns.likelihood_evaluations,
            logZ=self.ns.logZ_history,
            dZ=self.ns.dZ_history,
            mean_acceptance=self.ns.mean_acceptance_history,
            rolling_p=self.ns.rolling_p,
            population=dict(
                iterations=self.ns.population_iterations,
                acceptance=self.ns.population_acceptance,
            ),
            training_iterations=self.ns.training_iterations,
        )
        d["insertion_indices"] = self.ns.insertion_indices
        d["nested_samples"] = live_points_to_dict(self.nested_samples)
        d["posterior_samples"] = live_points_to_dict(self.posterior_samples)
        d["log_evidence"] = self.ns.log_evidence
        d["information"] = self.ns.information
        d["sampling_time"] = self.ns.sampling_time.total_seconds()
        d["training_time"] = self.ns.training_time.total_seconds()
        d["population_time"] = self.ns.proposal_population_time.total_seconds()
        if self.ns.likelihood_evaluation_time.total_seconds():
            d[
                "likelihood_evaluation_time"
            ] = self.ns.likelihood_evaluation_time.total_seconds()

        with open(filename, "w") as wf:
            json.dump(d, wf, indent=4, cls=NessaiJSONEncoder)

    def safe_exit(self, signum=None, frame=None):
        """
        Safely exit. This includes closing the multiprocessing pool.
        """
        logger.warning(f"Trying to safely exit with code {signum}")
        self.ns.model.close_pool(code=signum)
        self.ns.checkpoint()

        logger.warning(f"Exiting with code: {self.exit_code}")
        sys.exit(self.exit_code)



# -*- coding: utf-8 -*-
"""
Functions and objects related to reparametersiations for use in the
``reparameterisations`` dictionary.

See the documentation for an in-depth description of how to use these
functions and classes.
"""
import logging

import numpy as np
from scipy import stats

from .utils import (
    detect_edge,
    configure_edge_detection,
    rescale_zero_to_one,
    inverse_rescale_zero_to_one,
    rescale_minus_one_to_one,
    inverse_rescale_minus_one_to_one,
    determine_rescaled_bounds,
    rescaling_functions,
)

from .priors import (
    log_uniform_prior,
    log_2d_cartesian_prior,
    log_2d_cartesian_prior_sine,
)


logger = logging.getLogger(__name__)


def get_reparameterisation(reparameterisation, defaults=None):
    """Function to get a reparameterisation class from a name

    Parameters
    ----------
    reparameterisation : str, \
            :obj:`nessai.reparameterisations.Reparameterisation`
        Name of the reparameterisations to return or a class that inherits from
        :obj:`~nessai.reparameterisations.Reparameterisation`
    defaults : dict, optional
        Dictionary of known reparameterisations that overrides the defaults.

    Returns
    -------
    :obj:`nessai.reparameteristaions.Reparameterisation`
        Reparameterisation class.
    dict
        Keyword arguments for the specific reparameterisation.
    """
    if defaults is None:
        defaults = default_reparameterisations
    if isinstance(reparameterisation, str):
        rc, kwargs = defaults.get(reparameterisation, (None, None))
        if rc is None:
            raise ValueError(
                f"Unknown reparameterisation: {reparameterisation}"
            )
        else:
            if kwargs is None:
                kwargs = {}
            else:
                kwargs = kwargs.copy()
            return rc, kwargs
    elif reparameterisation is None:
        return NullReparameterisation, {}
    elif isinstance(reparameterisation, type) and issubclass(
        reparameterisation, Reparameterisation
    ):
        return reparameterisation, {}
    else:
        raise TypeError(
            "Reparameterisation must be a str, None, or class that "
            "inherits from `Reparameterisation`"
        )


class Reparameterisation:
    """
    Base object for reparameterisations.

    Parameters
    ----------
    parameters : str or list
        Name of parameters to reparameterise.
    prior_bounds : list, dict or None
        Prior bounds for the parameter(s).
    """

    _update_bounds = False
    has_prior = False
    has_prime_prior = False
    requires_prime_prior = False
    requires_bounded_prior = False
    prior_bounds = None
    prime_prior_bounds = None

    def __init__(self, parameters=None, prior_bounds=None):
        if not isinstance(parameters, (str, list)):
            raise TypeError("Parameters must be a str or list.")

        self.parameters = (
            [parameters] if isinstance(parameters, str) else parameters.copy()
        )

        if isinstance(prior_bounds, (list, tuple, np.ndarray)):
            if len(prior_bounds) == 2:
                prior_bounds = {self.parameters[0]: np.asarray(prior_bounds)}
            else:
                raise RuntimeError("Prior bounds got a list of len > 2")
        elif prior_bounds is None:
            if self.requires_bounded_prior:
                raise RuntimeError(
                    f"Reparameterisation {self.name} requires prior bounds!"
                )
            else:
                self.prior_bounds = None
        elif not isinstance(prior_bounds, dict):
            raise TypeError(
                "Prior bounds must be a dict, tuple, list or numpy array"
                " of len 2 or None."
            )

        if prior_bounds is not None:
            if set(self.parameters) - set(prior_bounds.keys()):
                raise RuntimeError(
                    "Mismatch between parameters and prior bounds: "
                    f"{set(self.parameters)}, {set(prior_bounds.keys())}"
                )
            self.prior_bounds = {
                p: np.asarray(b) for p, b in prior_bounds.items()
            }
        else:
            logger.debug(f"No prior bounds for {self.name}")

        if self.requires_bounded_prior:
            is_finite = np.isfinite(
                [pb for pb in self.prior_bounds.values()]
            ).all()
            if not is_finite:
                raise RuntimeError(
                    f"Reparameterisation {self.name} requires finite prior "
                    f"bounds. Received: {self.prior_bounds}"
                )

        self.prime_parameters = [p + "_prime" for p in self.parameters]
        self.requires = []
        logger.debug(f"Initialised reparameterisation: {self.name}")

    @property
    def name(self):
        """Unique name of the reparameterisations"""
        return (
            self.__class__.__name__.lower() + "_" + "_".join(self.parameters)
        )

    def reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space to x'-space.

        Parameters
        ----------
        x : structured array
            Array of inputs
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError

    def inverse_reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError


class CombinedReparameterisation(dict):
    """Class to handle multiple reparameterisations

    Parameters
    ----------
    reparameterisations : list, optional
        List of reparameterisations to add to the combined reparameterisations.
        Further reparameterisations can be added using
        `add_reparameterisations`.
    """

    def __init__(self, reparameterisations=None):
        super().__init__()
        self.reparameterisations = {}
        self.parameters = []
        self.prime_parameters = []
        self.requires = []
        if reparameterisations is not None:
            self.add_reparameterisations(reparameterisations)

    @property
    def has_prime_prior(self):
        """Boolean to check if prime prior can be enabled"""
        return all(r.has_prime_prior for r in self.values())

    @property
    def requires_prime_prior(self):
        """Boolean to check if any of the priors require the prime space"""
        return any(r.requires_prime_prior for r in self.values())

    def _add_reparameterisation(self, reparameterisation):
        requires = reparameterisation.requires
        if requires and (
            requires not in self.parameters
            or requires not in self.prime_parameters
        ):
            raise RuntimeError(
                f"Could not add {reparameterisation}, missing requirement(s): "
                f"{reparameterisation.requires}."
            )

        self[reparameterisation.name] = reparameterisation
        self.parameters += reparameterisation.parameters
        self.prime_parameters += reparameterisation.prime_parameters
        self.requires += reparameterisation.requires

    def add_reparameterisation(self, reparameterisation):
        """Add a reparameterisation"""
        self.add_reparameterisations(reparameterisation)

    def add_reparameterisations(self, reparameterisations):
        """Add multiple reparameterisations

        Parameters
        ----------
        reparameterisations : list of :`obj`:Reparameterisation
            List of reparameterisations to add.
        """
        if not isinstance(reparameterisations, list):
            reparameterisations = [reparameterisations]
        for r in reparameterisations:
            self._add_reparameterisation(r)

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for r in self.values():
            x, x_prime, log_j = r.reparameterise(x, x_prime, log_j, **kwargs)
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for r in reversed(list(self.values())):
            x, x_prime, log_j = r.inverse_reparameterise(
                x, x_prime, log_j, **kwargs
            )
        return x, x_prime, log_j

    def update_bounds(self, x):
        """
        Update the bounds used for the reparameterisation
        """
        for r in self.values():
            if hasattr(r, "update_bounds"):
                logger.debug(f"Updating bounds for: {r.name}")
                r.update_bounds(x)

    def reset_inversion(self):
        """
        Reset edges for boundary inversion
        """
        for r in self.values():
            if hasattr(r, "reset_inversion"):
                r.reset_inversion()

    def log_prior(self, x):
        """
        Compute any additional priors for auxiliary parameters
        """
        log_p = 0
        for r in self.values():
            if r.has_prior:
                log_p += r.log_prior(x)
        return log_p

    def x_prime_log_prior(self, x_prime):
        """
        Compute the prior in the prime space
        """
        log_p = 0
        for r in self.values():
            log_p += r.x_prime_log_prior(x_prime)
        return log_p


class NullReparameterisation(Reparameterisation):
    """Reparameterisation that does not modify the parameters.

    Parameters
    ----------
    parameters : list or str
        Parameters for which the reparameterisation will be used.
    prior_bounds : list, dict or None
        Prior bounds for parameters. Unused for this reparameterisation.
    """

    def __init__(self, parameters=None, prior_bounds=None):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)
        self.prime_parameters = self.parameters
        logger.debug(f"Initialised reparameterisation: {self.name}")

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x_prime[self.prime_parameters] = x[self.parameters]
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x[self.parameters] = x_prime[self.prime_parameters]
        return x, x_prime, log_j


class Rescale(Reparameterisation):
    """Reparameterisation that rescales the parameters by a constant factor
    that does not depend on the prior bounds.

    Parameters
    ----------
    scale : float
        Scaling constant.
    """

    requires_bounded_prior = False

    def __init__(self, parameters=None, scale=None, prior_bounds=None):
        if scale is None:
            raise RuntimeError("Must specify a scale!")
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        if isinstance(scale, (int, float)):
            self.scale = {p: float(scale) for p in self.parameters}
        elif isinstance(scale, list):
            if not len(scale) == len(self.parameters):
                raise RuntimeError(
                    "Scale list is a different length to the parameters."
                )
            self.scale = {p: float(s) for p, s in zip(self.parameters, scale)}
        elif isinstance(scale, dict):
            if not set(self.parameters) == set(scale.keys()):
                raise RuntimeError("Mismatched parameters with scale dict")
            scale = {p: float(s) for p, s in scale.items()}
            self.scale = scale
        else:
            raise TypeError(
                "Scale input must be an instance of int, list or dict"
            )

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            x_prime[pp] = x[p] / self.scale[p]
            log_j -= np.log(np.abs(self.scale[p]))
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            x[p] = x_prime[pp] * self.scale[p]
            log_j += np.log(np.abs(self.scale[p]))
        return x, x_prime, log_j


class RescaleToBounds(Reparameterisation):
    """Reparameterisation that maps to the specified interval.

    By default the interval is [-1, 1]. Also includes options for
    boundary inversion.

    This reparameterisation can handle multiple parameters.

    Parameters
    ----------
    parameters : list of str
        List of the names of parameters
    prior_bounds : dict
        Dictionary of prior bounds for each parameter. Does not need to be
        specified by the user.
    rescale_bounds : list of tuples, optional
        Bounds to rescale to.
    update_bounds : bool, optional
        Enable or disable updating bounds.
    prior : {'uniform', None}
        Type of prior used, if uniform prime prior is enabled.
    boundary_inversion : bool, list, dict, optional
        Configuration for boundary inversion. If a list, inversion is only
        applied to the parameters in the list based on `inversion_type`. If
        a `dict` then each item should be a parameter and a corresponding
        inversion type `{'split', 'inversion'}`.
    detect_edges : bool, optional
        Enable or disable edge detection for inversion.
    detect_edges_kwargs : dict, optional
        Dictionary of kwargs used to configure edge detection.
    offset : bool, optional
        Enable or disable offset subtraction. If `True` then the mean value
        of the prior is subtract of the parameter before the rescaling is
        applied. This is computed and applied after the 'pre-rescaling' if it
        has been specified.
    pre_rescaling : tuple of functions
        A function that applies a rescaling prior to the main rescaling and
        its inverse. Each function should return a value and the log-Jacobian
        determinant.
    post_rescaling : tuple of functions or {'logit}
        A function that applies a rescaling after to the main rescaling and
        its inverse. Each function should return a value and the log-Jacobian
        determinant. For example applying a logit after rescaling to [0, 1].
    """

    requires_bounded_prior = True

    def __init__(
        self,
        parameters=None,
        prior_bounds=None,
        prior=None,
        rescale_bounds=None,
        boundary_inversion=None,
        detect_edges=False,
        inversion_type="split",
        detect_edges_kwargs=None,
        offset=False,
        update_bounds=True,
        pre_rescaling=None,
        post_rescaling=None,
    ):

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        self.bounds = None
        self.detect_edge_prime = False

        self.has_pre_rescaling = True
        self.has_post_rescaling = True

        if rescale_bounds is None:
            logger.debug("Using default rescale bounds: [-1, 1]")
            self.rescale_bounds = {p: [-1, 1] for p in self.parameters}
        else:
            if isinstance(rescale_bounds, list):
                self.rescale_bounds = {
                    p: rescale_bounds for p in self.parameters
                }
            elif isinstance(rescale_bounds, dict):
                s = set(parameters) - set(rescale_bounds.keys())
                if s:
                    raise RuntimeError(
                        f"Missing rescale bounds for parameters: {s}"
                    )
                self.rescale_bounds = rescale_bounds
            else:
                raise TypeError(
                    "rescale_bounds must be an instance of list or dict. "
                    f"Got type: {type(rescale_bounds).__name__}"
                )

        if boundary_inversion is not None:
            if isinstance(boundary_inversion, list):
                self.boundary_inversion = {
                    n: inversion_type for n in boundary_inversion
                }
            elif isinstance(boundary_inversion, dict):
                self.boundary_inversion = boundary_inversion
            elif isinstance(boundary_inversion, bool):
                self.boundary_inversion = {
                    p: inversion_type for p in self.parameters
                }
            else:
                raise TypeError(
                    "boundary_inversion must be a list, dict or bool. "
                    f"Got type: {type(boundary_inversion).__name__}"
                )
        else:
            self.boundary_inversion = []

        for p in self.boundary_inversion:
            self.rescale_bounds[p] = [0, 1]

        self._update_bounds = update_bounds if not detect_edges else True
        self._edges = {n: None for n in self.parameters}
        self.detect_edges = detect_edges
        if self.boundary_inversion:
            self.detect_edges_kwargs = configure_edge_detection(
                detect_edges_kwargs, self.detect_edges
            )

        if self.detect_edges and not self.boundary_inversion:
            raise RuntimeError(
                "Must enable boundary inversion to use detect edges"
            )

        if prior == "uniform":
            self.prior = "uniform"
            self.has_prime_prior = True
            self._prime_prior = log_uniform_prior
            logger.debug(f"Prime prior enabled for {self.name}")
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

        self.configure_pre_rescaling(pre_rescaling)
        self.configure_post_rescaling(post_rescaling)

        if offset:
            self.offsets = {
                p: self.pre_rescaling(b[0])[0]
                + np.ptp(self.pre_rescaling(b)[0]) / 2
                for p, b in self.prior_bounds.items()
            }
            logger.debug(f"Offsets: {self.offsets}")
        else:
            self.offsets = {p: 0.0 for p in self.prior_bounds.keys()}

        self.set_bounds(self.prior_bounds)

    def configure_pre_rescaling(self, pre_rescaling):
        """Configure the rescaling applied before the standard rescaling.

        Used in :code:`DistanceReparameterisation`.

        Parameters
        ----------
        pre_rescaling : str or Tuple[Callable, Callable]
            Name of the pre-rescaling of tuple contain the forward and inverse
            functions that should return the rescaled value and the Jacobian.
        """
        if pre_rescaling is not None:
            if isinstance(pre_rescaling, str):
                logger.debug(f"Getting pre-rescaling function {pre_rescaling}")
                (
                    self.pre_rescaling,
                    self.pre_rescaling_inv,
                ) = rescaling_functions.get(
                    pre_rescaling.lower(), (None, None)
                )
                if self.pre_rescaling is None:
                    raise RuntimeError(
                        f"Unknown rescaling function: {pre_rescaling}"
                    )
            elif len(pre_rescaling) == 2:
                self.pre_rescaling = pre_rescaling[0]
                self.pre_rescaling_inv = pre_rescaling[1]
            else:
                raise RuntimeError(
                    "Pre-rescaling must be a str or tuple of two functions"
                )
            self.has_pre_rescaling = True
        else:
            logger.debug("No pre-rescaling to configure")
            self.has_pre_rescaling = False

    def configure_post_rescaling(self, post_rescaling):
        """Configure the rescaling applied after the standard rescaling.

        Used to apply the logit/sigmoid transforms after rescaling to [0, 1]

        Parameters
        ----------
        post_rescaling : str or Tuple[Callable, Callable]
            Name of the post-rescaling of tuple contain the forward and inverse
            functions that should return the rescaled value and the Jacobian.
        """
        if post_rescaling is not None:
            if isinstance(post_rescaling, str):
                logger.debug(
                    f"Getting post-rescaling function {post_rescaling}"
                )
                (
                    self.post_rescaling,
                    self.post_rescaling_inv,
                ) = rescaling_functions.get(
                    post_rescaling.lower(), (None, None)
                )
                if self.post_rescaling is None:
                    raise RuntimeError(
                        f"Unknown rescaling function: {post_rescaling}"
                    )
            elif len(post_rescaling) == 2:
                self.post_rescaling = post_rescaling[0]
                self.post_rescaling_inv = post_rescaling[1]
            else:
                raise RuntimeError(
                    "Post-rescaling must be a str or tuple of two functions"
                )
            logger.debug("Disabling prime prior with post-rescaling")
            self.has_prime_prior = False

            if post_rescaling == "logit":
                if self._update_bounds:
                    raise RuntimeError("Cannot use logit with update bounds")
                logger.debug("Setting bounds to [0, 1] for logit")
                self.rescale_bounds = {p: [0, 1] for p in self.parameters}
            self.has_post_rescaling = True
        else:
            logger.debug("No post-rescaling to configure")
            self.has_post_rescaling = False

    def pre_rescaling(self, x):
        """Function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def pre_rescaling_inv(self, x):
        """Inverse of function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def post_rescaling(self, x):
        """Function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def post_rescaling_inv(self, x):
        """Inverse of function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def _rescale_to_bounds(self, x, n):
        out = (
            self._rescale_factor[n]
            * (
                (x - self.bounds[n][0])
                / (self.bounds[n][1] - self.bounds[n][0])
            )
            + self._rescale_shift[n]
        )

        log_j = -np.log(self.bounds[n][1] - self.bounds[n][0]) + np.log(
            self._rescale_factor[n]
        )
        return out, log_j

    def _inverse_rescale_to_bounds(self, x, n):
        out = (self.bounds[n][1] - self.bounds[n][0]) * (
            x - self._rescale_shift[n]
        ) / self._rescale_factor[n] + self.bounds[n][0]

        log_j = np.log(self.bounds[n][1] - self.bounds[n][0]) - np.log(
            self._rescale_factor[n]
        )

        return out, log_j

    def _apply_inversion(
        self, x, x_prime, log_j, p, pp, compute_radius, test=None
    ):
        if self._edges[p] is None:
            self._edges[p] = detect_edge(
                x_prime[pp], test=test, **self.detect_edges_kwargs
            )
            self.update_prime_prior_bounds()

        if self._edges[p]:
            logger.debug(f"Apply inversion for {p} to {self._edges[p]} bound")
            logger.debug("Fixing bounds to [0, 1]")
            logger.debug("Rescaling")
            x_prime[pp], lj = rescale_zero_to_one(
                x_prime[pp] - self.offsets[p], *self.bounds[p]
            )
            log_j += lj
            if self._edges[p] == "upper":
                x_prime[pp] = 1 - x_prime[pp]
            if self.boundary_inversion[p] == "duplicate" or compute_radius:
                logger.debug("Inverting with duplication")
                x_inv = x_prime.copy()
                x_inv[pp] *= -1
                x_prime = np.concatenate([x_prime, x_inv])
                x = np.concatenate([x, x])
                log_j = np.concatenate([log_j, log_j])
            else:
                logger.debug("Inverting with splitting")
                inv = np.random.choice(
                    x_prime.size, x_prime.size // 2, replace=False
                )
                x_prime[pp][inv] *= -1
        else:
            logger.debug(f"Not using inversion for {p}")
            logger.debug(f"Rescaling to {self.rescale_bounds[p]}")
            x_prime[pp], lj = rescale_minus_one_to_one(
                x_prime[pp] - self.offsets[p],
                xmin=self.bounds[p][0],
                xmax=self.bounds[p][1],
            )

            log_j += lj

        return x, x_prime, log_j

    def _reverse_inversion(self, x, x_prime, log_j, p, pp):
        if self._edges[p]:
            inv = x[p] < 0.0
            x[p][~inv] = x[p][~inv]
            x[p][inv] = -x[p][inv]

            if self._edges[p] == "upper":
                x[p] = 1 - x[p]
            x[p], lj = inverse_rescale_zero_to_one(x[p], *self.bounds[p])
            x[p] += self.offsets[p]
            log_j += lj
        else:
            x[p], lj = inverse_rescale_minus_one_to_one(
                x[p], xmin=self.bounds[p][0], xmax=self.bounds[p][1]
            )
            x[p] += self.offsets[p]
            log_j += lj
        return x, x_prime, log_j

    def reparameterise(
        self, x, x_prime, log_j, compute_radius=False, **kwargs
    ):
        """
        Rescale inputs to the prime space

        Parameters
        ----------
        x, x_prime :  array_like
            Arrays of samples in the physical and prime space
        log_j : array_like
            Array of values of log-Jacobian
        compute_radius : bool, optional
            If true force duplicate for inversion
        kwargs :
            Parsed to inversion function
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            if self.has_pre_rescaling:
                x_prime[pp], lj = self.pre_rescaling(x[p])
                log_j += lj
            else:
                x_prime[pp] = x[p].copy()

            if p in self.boundary_inversion:
                x, x_prime, log_j = self._apply_inversion(
                    x, x_prime, log_j, p, pp, compute_radius, **kwargs
                )
            else:
                x_prime[pp], lj = self._rescale_to_bounds(
                    x_prime[pp] - self.offsets[p], p
                )
                log_j += lj
            if self.has_post_rescaling:
                x_prime[pp], lj = self.post_rescaling(x_prime[pp])
                log_j += lj
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Map inputs to the physical space from the prime space"""
        for p, pp in zip(
            reversed(self.parameters), reversed(self.prime_parameters)
        ):
            if self.has_post_rescaling:
                x[p], lj = self.post_rescaling_inv(x_prime[pp])
                log_j += lj
            else:
                x[p] = x_prime[pp].copy()
            if p in self.boundary_inversion:
                x, x_prime, log_j = self._reverse_inversion(
                    x, x_prime, log_j, p, pp, **kwargs
                )
            else:
                x[p], lj = self._inverse_rescale_to_bounds(x[p], p)
                x[p] += self.offsets[p]
                log_j += lj
            if self.has_pre_rescaling:
                x[p], lj = self.pre_rescaling_inv(x[p])
                log_j += lj
        return x, x_prime, log_j

    def reset_inversion(self):
        """Reset the edges for inversion"""
        self._edges = {n: None for n in self.parameters}

    def set_bounds(self, prior_bounds):
        """Set the initial bounds for rescaling"""
        self._rescale_factor = {
            p: np.ptp(self.rescale_bounds[p]) for p in self.parameters
        }
        self._rescale_shift = {
            p: self.rescale_bounds[p][0] for p in self.parameters
        }

        self.pre_prior_bounds = {
            p: self.pre_rescaling(prior_bounds[p])[0] for p in self.parameters
        }
        self.bounds = {
            p: self.pre_rescaling(b)[0] - self.offsets[p]
            for p, b in prior_bounds.items()
        }
        logger.debug(f"Initial bounds: {self.bounds}")
        self.update_prime_prior_bounds()

    def update_bounds(self, x):
        """Update the bounds used for the reparameterisation"""
        if self._update_bounds:
            self.bounds = {
                p: [
                    self.pre_rescaling(np.min(x[p]))[0] - self.offsets[p],
                    self.pre_rescaling(np.max(x[p]))[0] - self.offsets[p],
                ]
                for p in self.parameters
            }
            logger.debug(f"New bounds: {self.bounds}")
            self.update_prime_prior_bounds()

        else:
            logger.debug("Update bounds not enabled")

    def update_prime_prior_bounds(self):
        """Update the prior bounds used for the prime prior"""
        if self.has_prime_prior:
            self.prime_prior_bounds = {
                pp: self.post_rescaling(
                    np.asarray(
                        determine_rescaled_bounds(
                            self.pre_prior_bounds[p][0],
                            self.pre_prior_bounds[p][1],
                            self.bounds[p][0],
                            self.bounds[p][1],
                            invert=self._edges[p],
                            inversion=p in self.boundary_inversion,
                            offset=self.offsets[p],
                            rescale_bounds=self.rescale_bounds[p],
                        )
                    )
                )[0]
                for p, pp in zip(self.parameters, self.prime_parameters)
            }
            logger.debug(f"New prime bounds: {self.prime_prior_bounds}")

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            log_p = 0
            for pp in self.prime_parameters:
                log_p += self._prime_prior(
                    x_prime[pp], *self.prime_prior_bounds[pp]
                )
            return log_p
        else:
            raise RuntimeError(
                f"Prime prior is not configured for {self.name}"
            )


class Angle(Reparameterisation):
    """Reparameterisation for a single angle.

    This reparameterisations converts an angle to Cartesian coordinates
    using either a corresponding radial parameter or an auxiliary radial
    parameter. When using the auxiliary parameter, samples are drawn from
    a chi-distribution with two degrees of freedom.

    Parameters
    ----------
    parameters : str or list
        Parameter(s) to use for the reparameterisation. Either just an angle
        or an angle and corresponding radial parameter
    prior_bounds : dict
        Dictionary of prior bounds. Does not need to be specified when defining
        `reparameterisations`.
    scale : float, optional
        Value used to rescale the angle before converting to Cartesian
        coordinates. If None the scale will be set to 2pi / prior_bounds.
    prior : {'uniform', 'sine', None}
        Type of prior being used for sampling this angle. If specified, the
        prime prior is enabled. If None then it is disabled.
    """

    requires_bounded_prior = True

    def __init__(
        self, parameters=None, prior_bounds=None, scale=1.0, prior=None
    ):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        if len(self.parameters) == 1:
            self.parameters.append(self.parameters[0] + "_radial")
            self.chi = stats.chi(2)
            self.has_prior = True
        elif len(self.parameters) == 2:
            self.chi = False
        else:
            raise RuntimeError("Too many parameters for Angle")

        if scale is None:
            logger.debug("Scale is None, using 2pi / prior_range")
            self.scale = 2.0 * np.pi / np.ptp(self.prior_bounds[self.angle])
        else:
            self.scale = scale

        if prior_bounds[self.angle][0] == 0:
            self._zero_bound = True
        else:
            self._zero_bound = False

        self.prime_parameters = [self.angle + "_x", self.angle + "_y"]
        self.requires = []

        if prior in ["uniform", "sine"]:
            self.prior = prior
            self.has_prime_prior = True
            if self.prior == "uniform":
                self._prime_prior = log_2d_cartesian_prior
                self._k = self.scale * np.pi
            else:
                self._prime_prior = log_2d_cartesian_prior_sine
                self._k = np.pi
            logger.debug(f"Prime prior enabled for {self.name}")
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

    @property
    def angle(self):
        """The name of the angular parameter"""
        return self.parameters[0]

    @property
    def radial(self):
        """The name of the radial parameter"""
        return self.parameters[1]

    @property
    def radius(self):
        """The name of the radial parameter (equivalent to radial)"""
        return self.parameters[1]

    @property
    def x(self):
        """The name of x coordinate"""
        return self.prime_parameters[0]

    @property
    def y(self):
        """The name of y coordinate"""
        return self.prime_parameters[1]

    def _rescale_radial(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[1]], x, x_prime, log_j

    def _rescale_angle(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[0]] * self.scale, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert the angle to Cartesian coordinates"""
        angle, x, x_prime, log_j = self._rescale_angle(
            x, x_prime, log_j, **kwargs
        )

        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r, x, x_prime, log_j = self._rescale_radial(
                x, x_prime, log_j, **kwargs
            )
        if any(r < 0):
            raise RuntimeError("Radius cannot be negative.")

        x_prime[self.prime_parameters[0]] = r * np.cos(angle)
        x_prime[self.prime_parameters[1]] = r * np.sin(angle)
        log_j += np.log(r)
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to an angle and radial component"""
        x[self.parameters[1]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2
            + x_prime[self.prime_parameters[1]] ** 2
        )
        if self._zero_bound:
            x[self.parameters[0]] = (
                np.arctan2(
                    x_prime[self.prime_parameters[1]],
                    x_prime[self.prime_parameters[0]],
                )
                % (2.0 * np.pi)
                / self.scale
            )
        else:
            x[self.parameters[0]] = (
                np.arctan2(
                    x_prime[self.prime_parameters[1]],
                    x_prime[self.prime_parameters[0]],
                )
                / self.scale
            )

        log_j -= np.log(x[self.parameters[1]])
        x, x_prime, log_j = self._inverse_rescale_angle(x, x_prime, log_j)

        return x, x_prime, log_j

    def log_prior(self, x):
        """Prior for radial parameter"""
        return self.chi.logpdf(x[self.parameters[1]])

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            return self._prime_prior(
                x_prime[self.prime_parameters[0]],
                x_prime[self.prime_parameters[1]],
                k=self._k,
            )
        else:
            raise RuntimeError("Prime prior")


class ToCartesian(Angle):
    """Convert a parameter to Cartesian coordinates"""

    def __init__(self, mode="split", scale=np.pi, **kwargs):
        super().__init__(scale=scale, **kwargs)

        self.mode = mode
        if self.mode not in ["duplicate", "split", "half"]:
            raise RuntimeError(f"Unknown mode: {self.mode}")
        logger.debug(f"Using mode: {self.mode}")

        self._zero_bound = False
        self._k = self.prior_bounds[self.parameters[0]][1]

    def _rescale_angle(
        self, x, x_prime, log_j, compute_radius=False, **kwargs
    ):
        angle, lj = rescale_zero_to_one(
            x[self.parameters[0]], *self.prior_bounds[self.parameters[0]]
        )
        log_j += lj
        if self.mode == "duplicate" or compute_radius:
            angle = np.concatenate([angle, -angle])
            x = np.concatenate([x, x])
            x_prime = np.concatenate([x_prime, x_prime])
            log_j = np.concatenate([log_j, log_j])
        elif self.mode == "split":
            neg = np.random.choice(angle.size, angle.size // 2, replace=False)
            angle[neg] *= -1

        angle *= self.scale
        return angle, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        x[self.parameters[0]], lj = inverse_rescale_zero_to_one(
            np.abs(x[self.parameters[0]]),
            *self.prior_bounds[self.parameters[0]],
        )
        log_j += lj
        return x, x_prime, log_j


class AnglePair(Reparameterisation):
    """Reparameterisation for a pair of angles and a radial component.

    Converts to three-dimensional Cartesian coordinates.

    If the radial component is not specified, it is sampled from a chi-
    distribution with three degrees of freedom.

    Notes
    -----
    The parameters will be reordered such that the first parameter is the angle
    along the horizon, the second parameter is the vertical angle and the last
    parameter is the radial parameter.

    Parameters
    -----------
    parameters : list
        List of parameters. Must contain at least the two angles and,
        optionally, also a radial component.
    prior_bounds : dict
        Dictionary of prior bounds for each parameter
    prior : str, {'isotropic', None}
        Type of prior, used to enable use of the prime prior.
    convention : str, {'ra-dec', 'az-zen'}
        Convention used for defining the spherical polar coordinates. If not
        set, it will be guessed based on either dec or zen. Where it is assumed
        declination is defined on [-pi/2, pi/2] and zenith on [0, pi].
    """

    requires_bounded_prior = True
    known_priors = ["isotropic", None]
    _conventions = {"az-zen": [0, np.pi], "ra-dec": [-np.pi / 2, np.pi / 2]}

    def __init__(
        self, parameters=None, prior_bounds=None, prior=None, convention=None
    ):

        if len(parameters) not in [2, 3]:
            raise RuntimeError(
                "Must use a pair of angles or a pair plus a radius"
            )

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        # Determine which parameter is for the horizontal plane
        # and which is for the vertical plane
        logger.debug("Checking order of parameters")
        b = np.ptp([prior_bounds[p] for p in parameters], axis=1)
        try:
            hz = np.where(b == (2 * np.pi))[0][0]
            vt = np.where(b == np.pi)[0][0]
        except IndexError:
            raise ValueError(
                f"Invalid prior ranges: {self.prior_bounds}. "
                "Parameters must be defined over a range of pi and 2 pi!"
            )
        # Make sure order is horizon, vertical, radial
        if len(parameters) == 3:
            r = list({0, 1, 2} - {hz, vt})[0]
            parameters[0], parameters[1], parameters[2] = (
                parameters[hz],
                parameters[vt],
                parameters[r],
            )
            m = "_".join(parameters)
            self.chi = False
        else:
            parameters[0], parameters[1] = parameters[hz], parameters[vt]
            m = "_".join(parameters)
            parameters.append(f"{m}_radial")
            self.chi = stats.chi(3)
            self.has_prior = True

        logger.debug(f"Parameters are: {parameters}")

        # Modulo 2pi is used if the first parameter (ra/az) in defined on
        # [0, 2pi] since output would otherwise be [-pi, pi]
        if np.array_equal(self.prior_bounds[parameters[0]], [0, 2 * np.pi]):
            self._modulo_2pi = True
        elif np.array_equal(self.prior_bounds[parameters[0]], [-np.pi, np.pi]):
            self._modulo_2pi = False
        else:
            raise ValueError(
                f"Prior bounds for {parameters[0]} must be [0, 2pi] or "
                f"[-pi, pi]. Received: {self.prior_bounds[parameters[0]]}"
            )

        self.parameters = parameters
        self.prime_parameters = [f"{m}_{x}" for x in ["x", "y", "z"]]

        if prior == "isotropic" and self.chi:
            self.has_prime_prior = True
            logger.info(f"Prime prior enabled for {self.name}")
        elif prior not in self.known_priors:
            raise ValueError(
                f"Unknown prior: `{prior}`. Choose from: {self.known_priors}"
            )
        else:
            self.has_prime_prior = False
            logger.debug(f"Prime prior disabled for {self.name}")

        if convention is None:
            logger.debug("Trying to determine convention")
            if np.array_equal(
                self.prior_bounds[self.parameters[1]], [0, np.pi]
            ):
                self.convention = "az-zen"
            elif np.array_equal(
                self.prior_bounds[self.parameters[1]], [-np.pi / 2, np.pi / 2]
            ):
                self.convention = "ra-dec"
            else:
                raise RuntimeError(
                    f"Could not determine convention for: {self.parameters}!"
                )
        elif convention in self._conventions.keys():
            self.convention = convention
            if not np.array_equal(
                self.prior_bounds[parameters[1]], self._conventions[convention]
            ):
                raise ValueError(
                    f"Prior bounds for {parameters[1]} must be "
                    f"{self._conventions[convention]} for the "
                    f"{convention} convention. "
                    f"Received: {self.prior_bounds[parameters[1]]}."
                )
        else:
            raise ValueError(
                f"Unknown convention: `{convention}`. "
                f"Choose from: {list(self._conventions.keys())}."
            )

        logger.debug(f"Using convention: {self.convention}")

        self.requires = []

    @property
    def angles(self):
        """Names of the two angles.

        Order is: angle along the horizon, vertical angle.
        """
        return self.parameters[:2]

    @property
    def radial(self):
        """Name of the radial parameter"""
        return self.parameters[-1]

    @property
    def x(self):
        """Name of the first Cartesian coordinate"""
        return self.prime_parameters[0]

    @property
    def y(self):
        """Name of the second Cartesian coordinate"""
        return self.prime_parameters[1]

    @property
    def z(self):
        """Name of the third Cartesian coordinate"""
        return self.prime_parameters[2]

    def _az_zen(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = (
            r * np.sin(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[1]] = (
            r * np.sin(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[2]] = r * np.cos(x[self.parameters[1]])
        log_j += 2 * np.log(r) + np.log(np.sin(x[self.parameters[1]]))
        return x, x_prime, log_j

    def _ra_dec(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = (
            r * np.cos(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[1]] = (
            r * np.cos(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        )
        x_prime[self.prime_parameters[2]] = r * np.sin(x[self.parameters[1]])
        log_j += 2 * np.log(r) + np.log(np.cos(x[self.parameters[1]]))
        return x, x_prime, log_j

    def _inv_az_zen(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2.0
            + x_prime[self.prime_parameters[1]] ** 2.0
            + x_prime[self.prime_parameters[2]] ** 2.0
        )

        if self._modulo_2pi:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            ) % (2.0 * np.pi)
        else:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            )
        x[self.parameters[1]] = np.arctan2(
            np.sqrt(
                x_prime[self.prime_parameters[0]] ** 2.0
                + x_prime[self.prime_parameters[1]] ** 2.0
            ),
            x_prime[self.prime_parameters[2]],
        )
        log_j += -2 * np.log(x[self.parameters[2]]) - np.log(
            np.sin(x[self.parameters[1]])
        )

        return x, x_prime, log_j

    def _inv_ra_dec(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2.0
            + x_prime[self.prime_parameters[1]] ** 2.0
            + x_prime[self.prime_parameters[2]] ** 2.0
        )

        if self._modulo_2pi:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            ) % (2.0 * np.pi)
        else:
            x[self.parameters[0]] = np.arctan2(
                x_prime[self.prime_parameters[1]],
                x_prime[self.prime_parameters[0]],
            )

        x[self.parameters[1]] = np.arctan2(
            x_prime[self.prime_parameters[2]],
            np.sqrt(
                x_prime[self.prime_parameters[0]] ** 2.0
                + x_prime[self.prime_parameters[1]] ** 2.0
            ),
        )

        log_j += -2 * np.log(x[self.parameters[2]]) - np.log(
            np.cos(x[self.parameters[1]])
        )
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Convert the spherical polar angles to Cartesian coordinates
        """
        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r = x[self.radial]
        if any(r < 0):
            raise RuntimeError("Radius cannot be negative.")

        if self.convention == "az-zen":
            return self._az_zen(x, x_prime, log_j, r)
        else:
            return self._ra_dec(x, x_prime, log_j, r)

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to spherical polar angles"""
        if self.convention == "az-zen":
            return self._inv_az_zen(x, x_prime, log_j)
        else:
            return self._inv_ra_dec(x, x_prime, log_j)

    def log_prior(self, x):
        """Prior for radial parameter"""
        if self.chi and self.has_prior:
            return self.chi.logpdf(x[self.parameters[2]])
        else:
            raise RuntimeError(
                "log_prior is not defined when a radial parameter has been "
                "specified!"
            )

    def x_prime_log_prior(self, x_prime):
        """
        Log probability of 3d Cartesian coordinates for an isotropic
        distribution of angles and a radial component drawn from a chi
        distribution with three degrees of freedom.
        """
        if self.has_prime_prior:
            return (
                -1.5 * np.log(2 * np.pi)
                - np.sum(
                    [x_prime[pp] ** 2 for pp in self.prime_parameters], axis=0
                )
                / 2
            )
        else:
            raise RuntimeError("x prime prior is not defined!")


default_reparameterisations = {
    "default": (RescaleToBounds, None),
    "rescaletobounds": (RescaleToBounds, None),
    "rescale-to-bounds": (RescaleToBounds, None),
    "offset": (RescaleToBounds, {"offset": True}),
    "inversion": (
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "split",
        },
    ),
    "inversion-duplicate": (
        RescaleToBounds,
        {
            "detect_edges": True,
            "boundary_inversion": True,
            "inversion_type": "duplicate",
        },
    ),
    "logit": (
        RescaleToBounds,
        {
            "rescale_bounds": [0.0, 1.0],
            "update_bounds": False,
            "post_rescaling": "logit",
        },
    ),
    "scale": (Rescale, None),
    "rescale": (Rescale, None),
    "angle": (Angle, {}),
    "angle-pi": (Angle, {"scale": 2.0, "prior": "uniform"}),
    "angle-2pi": (Angle, {"scale": 1.0, "prior": "uniform"}),
    "angle-sine": (RescaleToBounds, None),
    "angle-cosine": (RescaleToBounds, None),
    "angle-pair": (AnglePair, None),
    "periodic": (Angle, {"scale": None}),
    "to-cartesian": (ToCartesian, None),
    "none": (NullReparameterisation, None),
    "null": (NullReparameterisation, None),
    None: (NullReparameterisation, None),
}



# -*- coding: utf-8 -*-
"""
Test the base flow class
"""
import pytest
from unittest.mock import create_autospec, patch

from nessai.flows.base import BaseFlow, NFlow
from nflows.transforms import Transform


@pytest.fixture
def flow():
    return create_autospec(BaseFlow)


def test_base_flow_abstract_methods():
    """Assert an error is raised if the methods are not implemented."""
    with pytest.raises(TypeError) as excinfo:
        BaseFlow()
    assert "instantiate abstract class BaseFlow with abstract method" in str(
        excinfo.value
    )


@pytest.mark.parametrize(
    "method",
    [
        "forward",
        "inverse",
        "sample",
        "log_prob",
        "base_distribution_log_prob",
        "forward_and_log_prob",
        "sample_and_log_prob",
    ],
)
def test_base_flow_methods(method, flow):
    """Test to make sure all of the method in the base class raise errors"""
    with pytest.raises(NotImplementedError):
        getattr(BaseFlow, method)(flow, None)


def test_base_flow_to(flow):
    """Assert the to method works as intended"""
    flow.device = None
    with patch("nessai.flows.base.Module.to") as super_init:
        BaseFlow.to(flow, "cpu")
    super_init.assert_called_once_with("cpu")
    assert flow.device == "cpu"


def test_nflow_base_transform():
    """Assert an error is raised if the transform class is invalid."""

    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Test(), Test())
    assert "transform must inherit" in str(excinfo.value)


def test_nflow_base_distribution():
    """Assert an error is raised if the distribution class is invalid."""

    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Transform(), Test())
    assert "distribution must inherit" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
Test the FlowModel object.
"""
import numpy as np
import pickle
import pytest
import torch
from unittest.mock import create_autospec, MagicMock, patch

from nessai.flowmodel import update_config, FlowModel


@pytest.fixture()
def data_dim():
    return 2


@pytest.fixture()
def model():
    return create_autospec(FlowModel)


@pytest.fixture(scope="function")
def flow_model(flow_config, data_dim, tmpdir):
    flow_config["model_config"]["n_inputs"] = data_dim
    output = str(tmpdir.mkdir("flowmodel"))
    return FlowModel(flow_config, output=output)


def test_update_config_none():
    """Test update config for no input"""
    config = update_config(None)
    assert "model_config" in config


def test_update_config_invalid_type():
    """Test update config when an invalid argument is specified"""
    with pytest.raises(TypeError) as excinfo:
        update_config(False)
    assert "Must pass a dictionary" in str(excinfo.value)


@pytest.mark.parametrize("noise_scale", ["auto", 4])
def test_update_config_invalid_noise_scale(noise_scale):
    """Assert an error is raised if noise_scale is not a float or adapative."""
    config = {"noise_scale": noise_scale}
    with pytest.raises(ValueError) as excinfo:
        update_config(config)
    assert "noise_scale must be a float or" in str(excinfo.value)


def test_update_config_n_neurons():
    """Assert the n_neurons is set to 2x n_inputs"""
    config = dict(model_config=dict(n_inputs=10))
    config = update_config(config)
    assert config["model_config"]["n_neurons"] == 20


def test_init_no_config(tmpdir):
    """Test the init method with no config specified"""
    output = str(tmpdir.mkdir("no_config"))
    default_config = update_config(None)
    fm = FlowModel(output=output)

    assert fm.model_config == default_config["model_config"]


def test_init_config_class(tmpdir):
    """Test the init and save methods when specifying `flow` as a class"""
    from nessai.flows import RealNVP

    output = str(tmpdir.mkdir("no_config"))
    config = dict(model_config=dict(flow=RealNVP))
    fm = FlowModel(config=config, output=output)

    assert fm.model_config["flow"].__name__ == "RealNVP"


def test_initialise(model):
    """Test the initialise method"""
    model.get_optimiser = MagicMock()
    model.model_config = dict(n_neurons=2)
    model.inference_device = None
    model.optimiser = "adam"
    model.optimiser_kwargs = {"weights": 0.1}
    with patch(
        "nessai.flowmodel.configure_model", return_value=("model", "cpu")
    ) as mock:
        FlowModel.initialise(model)
    mock.assert_called_once_with(model.model_config)
    model.get_optimiser.assert_called_once_with("adam", weights=0.1)
    assert model.inference_device == torch.device("cpu")


@pytest.mark.parametrize("optimiser", ["Adam", "AdamW", "SGD"])
def test_get_optimiser(model, optimiser):
    """Test to make sure the three supported optimisers work"""
    model.lr = 0.1
    model.model = MagicMock()
    with patch(f"torch.optim.{optimiser}") as mock:
        FlowModel.get_optimiser(model, optimiser)
    mock.assert_called_once()


@pytest.mark.parametrize("val_size, batch_size", [(0.1, 100), (0.5, "all")])
def test_prep_data(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val = flow_model.prep_data(x, val_size, batch_size)
    if batch_size == "all":
        batch_size = int(n * (1 - val_size))

    assert flow_model.batch_size == batch_size
    assert len(train) + len(val) == n


@pytest.mark.parametrize("val_size, batch_size", [(0.1, 100), (0.5, "all")])
def test_prep_data_dataloader(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val = flow_model.prep_data(
        x, val_size, batch_size, use_dataloader=True
    )
    train_batch = next(iter(train))[0]
    val_batch = next(iter(val))[0]
    if batch_size == "all":
        batch_size = int(n * (1 - val_size))

    assert train.batch_size == batch_size
    assert list(val_batch.shape) == [int(val_size * n), data_dim]
    assert len(train) * train_batch.shape[0] + val_batch.shape[0] == n


@pytest.mark.parametrize("batch_size", [None, "10", True, False])
def test_incorrect_batch_size_type(flow_model, data_dim, batch_size):
    """Ensure the non-interger batch sizes do not work"""
    n = 1000
    x = np.random.randn(n, data_dim)
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.prep_data(x, 0.1, batch_size)
    assert "Unknown batch size" in str(excinfo.value)


@pytest.mark.parametrize("dataloader", [False, True])
def test_training(flow_model, data_dim, dataloader):
    """Test class until training"""
    x = np.random.randn(1000, data_dim)
    flow_model.use_dataloader = dataloader
    flow_model.train(x)
    assert flow_model.weights_file is not None


@pytest.mark.parametrize(
    "key, value",
    [
        ("annealing", True),
        ("noise_scale", 0.1),
        ("noise_scale", "adaptive"),
        ("max_epochs", 51),
    ],
)
def test_training_additional_config_args(
    flow_config, data_dim, tmpdir, key, value
):
    """
    Test training with different config args
    """
    flow_config["model_config"]["n_inputs"] = data_dim
    flow_config[key] = value

    output = str(tmpdir.mkdir("flowmodel"))
    flow_model = FlowModel(flow_config, output=output)

    assert getattr(flow_model, key) == value

    x = np.random.randn(100, data_dim)
    flow_model.train(x)


def test_early_optimiser_init(flow_model):
    """Ensure calling the opitmiser before the model raises an error"""
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.get_optimiser()
    assert "Cannot initialise optimiser" in str(excinfo.value)


@pytest.mark.parametrize("weights", [False, True])
@pytest.mark.parametrize("perms", [False, True])
def test_reset_model(model, weights, perms):
    """Test resetting the model"""
    model.model = MagicMock()
    model.apply = MagicMock()
    model.get_optimiser = MagicMock()
    model.optimiser = MagicMock()
    model.optimiser_kwargs = {"lr": 0.01}

    with patch(
        "nessai.flowmodel.configure_model",
        return_value=(MagicMock, "cpu"),
    ) as mock:
        FlowModel.reset_model(model, weights=weights, permutations=perms)

    if weights and perms:
        mock.assert_called_once()
    if any([weights, perms]):
        model.get_optimiser.assert_called_once_with(model.optimiser, lr=0.01)
    else:
        model.get_optimiser.assert_not_called()


def test_sample_and_log_prob_not_initialised(flow_model, data_dim):
    """
    Ensure user cannot call the method before the model initialise.
    """
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.sample_and_log_prob()
    assert "Model is not initialised" in str(excinfo.value)


@pytest.mark.parametrize("N", [1, 100])
def test_sample_and_log_prob_no_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob and ensure correct output
    shapes.
    """
    flow_model.initialise()
    x, log_prob = flow_model.sample_and_log_prob(N=N)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.parametrize("N", [1, 100])
def test_sample_and_log_prob_with_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob when samples from the
    latent space are provided
    """
    flow_model.initialise()
    z = np.random.randn(N, data_dim)
    x, log_prob = flow_model.sample_and_log_prob(z=z)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.parametrize("N", [1, 100])
def test_forward_and_log_prob(flow_model, data_dim, N):
    """Test the basic use of forward and log prob"""
    flow_model.initialise()
    x = np.random.randn(N, data_dim)
    z, log_prob = flow_model.forward_and_log_prob(x)
    assert z.shape == (N, data_dim)
    assert log_prob.size == N


def test_move_to_update_default(model):
    """Ensure the stored device is updated"""
    model.device = "cuda"
    model.model = MagicMock()
    model.model.to = MagicMock()
    FlowModel.move_to(model, "cpu", update_default=True)
    assert model.device == torch.device("cpu")
    model.model.to.assert_called_once()


@patch("torch.save")
def test_save_weights(mock_save, model):
    """Test saving the weights"""
    model.model = MagicMock()
    FlowModel.save_weights(model, "model.pt")
    mock_save.assert_called_once()
    assert model.weights_file == "model.pt"


def test_get_state(flow_model):
    """Make the object can be pickled"""
    pickle.dumps(flow_model)


@pytest.mark.integration_test
def test_lu_cache_reset(tmp_path):
    """Assert the LU cache is correctly reset after training.

    Cache is reset when calling .train() so if cache is incorrect then after
    the reset the values in the latent space will not match.
    """
    output = tmp_path / "test"
    output.mkdir()

    config = dict(
        max_epochs=100,
        patience=1000,
        model_config=dict(
            n_inputs=2,
            n_blocks=2,
            kwargs=dict(
                linear_transform="lu",
            ),
        ),
    )

    flow = FlowModel(config=config, output=output)
    data = np.random.randn(100, 2)

    flow.train(data)

    test_data = torch.from_numpy(data).type(torch.get_default_dtype())

    with torch.no_grad():
        z_out, log_j = flow.model.forward(test_data)
    flow.model.train()
    flow.model.eval()
    with torch.no_grad():
        z_out_reset, log_j_reset = flow.model.forward(test_data)

    np.testing.assert_array_equal(z_out_reset, z_out)
    np.testing.assert_array_equal(log_j_reset, log_j)



# -*- coding: utf-8 -*-
"""
Global configuration for nessai.
"""
import numpy as np

# Settings for live points
LOGL_DTYPE = "f8"
IT_DTYPE = "i4"
DEFAULT_FLOAT_DTYPE = "f8"
DEFAULT_FLOAT_VALUE = np.nan
CORE_PARAMETERS = ["logP", "logL", "it"]
CORE_PARAMETERS_DEFAULTS = [np.nan, np.nan, 0]
CORE_PARAMETERS_DTYPE = [DEFAULT_FLOAT_DTYPE, LOGL_DTYPE, IT_DTYPE]
EXTRA_PARAMETERS = []
EXTRA_PARAMETERS_DEFAULTS = []
EXTRA_PARAMETERS_DTYPE = []
NON_SAMPLING_PARAMETERS = CORE_PARAMETERS + EXTRA_PARAMETERS
NON_SAMPLING_DEFAULTS = CORE_PARAMETERS_DEFAULTS + EXTRA_PARAMETERS_DEFAULTS
NON_SAMPLING_DEFAULT_DTYPE = CORE_PARAMETERS_DTYPE + EXTRA_PARAMETERS_DTYPE
# Plotting config
DISABLE_STYLE = False
"""Disable nessai's custom plotting style globally.

Useful since all plotting functions use the
:py:func:`~nessai.plot.nessai_style` decorator by default.
"""
SNS_STYLE = "ticks"
"""Default seaborn style."""
BASE_COLOUR = "#02979d"
"""Base colour for plots."""
HIGHLIGHT_COLOUR = "#f5b754"
"""Highlight colour for plots."""
LINE_COLOURS = ["#4575b4", "#d73027", "#fad117", "#ff8c00"]
"""Default line colours."""
LINE_STYLES = ["-", "--", ":", "-."]
"""Default line styles."""



# -*- coding: utf-8 -*-
"""
Test plotting in the nested sampler.
"""
import os
import numpy as np
import pytest
from unittest.mock import MagicMock, patch
from nessai.nestedsampler import NestedSampler


@pytest.mark.parametrize("track_gradients", [False, True])
@pytest.mark.parametrize("filename", [None, "test.png"])
def test_plot_state(sampler, tmpdir, filename, track_gradients):
    """Test making the state plot"""
    x = np.arange(10)
    sampler.min_likelihood = x
    sampler.max_likelihood = x
    sampler.iteration = 1003
    sampler.training_iterations = [256, 711]
    sampler.train_on_empty = False
    sampler.population_iterations = [256, 500, 711, 800]
    sampler.population_acceptance = 4 * [0.5]
    sampler.population_radii = 4 * [1.0]
    sampler.checkpoint_iterations = [600]
    sampler.likelihood_evaluations = x
    sampler.state = MagicMock()
    sampler.state.log_vols = np.linspace(0, -10, 1050)
    sampler.state.track_gradients = track_gradients
    sampler.state.gradients = np.arange(1050)
    sampler.logZ_history = x
    sampler.dZ_history = x
    sampler.mean_acceptance_history = x
    sampler.rolling_p = np.arange(4)

    if filename is not None:
        sampler.output = tmpdir.mkdir("test_plot_state")
        filename = os.path.join(sampler.output, filename)
    fig = NestedSampler.plot_state(sampler, filename)

    if filename is not None:
        assert os.path.exists(filename)
    else:
        assert fig is not None


@pytest.mark.parametrize("samples", [[], [1, 2, 3]])
@pytest.mark.parametrize("filename", [None, "trace.png"])
@patch("nessai.nestedsampler.plot_trace", return_value="fig")
def test_plot_trace(mock_plot, sampler, tmpdir, samples, filename):
    """Test the plot_trace method"""
    sampler.nested_samples = samples
    sampler.state = MagicMock()
    sampler.state.log_vols = [1, 2, 3, 4]
    sampler.output = os.getcwd()

    if filename is not None:
        sampler.output = tmpdir.mkdir("test_plot_trace")
        filename = os.path.join(sampler.output, filename)

    fig = NestedSampler.plot_trace(sampler, filename=filename)

    if not len(samples):
        mock_plot.assert_not_called()
        assert fig is None
    else:
        mock_plot.assert_called_once_with(
            [2, 3, 4], samples, filename=filename
        )
        assert fig == "fig"


@pytest.mark.parametrize("filename", [None, "trace.png"])
@patch("nessai.nestedsampler.plot_indices", return_value="fig")
def test_plot_insertion_indices(mock_plot, sampler, filename):
    """Test plotting the insetion indices"""
    nlive = 10
    indices = list(range(20))
    sampler.nlive = nlive
    sampler.insertion_indices = indices
    fig = NestedSampler.plot_insertion_indices(
        sampler, filename=filename, k=True
    )
    assert fig == "fig"
    mock_plot.assert_called_once_with(
        indices, nlive, filename=filename, k=True
    )



# -*- coding: utf-8 -*-
"""Tests for the transforms included for the flows"""
from nessai.flows.transforms import LULinear
import pytest
import torch


def test_LULinear_weight_inverse():
    """Test the patched weight inverse method.

    Based on the test included in nflows:
    https://github.com/bayesiains/nflows/blob/master/tests/transforms/lu_test.py#L12
    """
    transform = LULinear(2)
    lower, upper = transform._create_lower_upper()
    weight = lower @ upper
    weight_inverse = torch.inverse(weight)

    out = transform.weight_inverse()

    assert not torch.isnan(out).any()
    assert not torch.isinf(out).any()
    assert out.shape == torch.Size((2, 2))
    assert torch.equal(out, weight_inverse)


@pytest.mark.cuda
def test_LUlinear_CUDA():
    """Test to verify that CUDA works"""
    transform = LULinear(2)
    transform.to("cuda")
    transform.weight_inverse()
